"""Provide our own abstract syntax tree for contract transpilation."""
import abc
import ast
import enum
from typing import Sequence, Union, Generic, TypeVar, List, Optional, Iterator

from icontract import DBC

from aas_core_codegen import stringify
from aas_core_codegen.common import Identifier, assert_never

T = TypeVar("T")


class Node(abc.ABC):
    """Represent an abstract node of our syntax tree."""

    def __init__(self, original_node: ast.AST) -> None:
        """Initialize with the given values."""
        self.original_node = original_node

    @abc.abstractmethod
    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        raise NotImplementedError()

    @abc.abstractmethod
    def visit(self, visitor: "Visitor") -> None:
        """Accept the transformer."""
        raise NotImplementedError()

    def __str__(self) -> str:
        """Provide a human-readable representation of the instance."""
        return dump(self)


class Statement(Node):
    """Represent a statement in a program."""

    @abc.abstractmethod
    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        raise NotImplementedError()

    @abc.abstractmethod
    def visit(self, visitor: "Visitor") -> None:
        """Accept the transformer."""
        raise NotImplementedError()


class Expression(Node):
    """Represent an expression in our abstract syntax tree."""

    @abc.abstractmethod
    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        raise NotImplementedError()

    @abc.abstractmethod
    def visit(self, visitor: "Visitor") -> None:
        """Accept the transformer."""
        raise NotImplementedError()


class Member(Expression):
    """
    Represent a member of an instance.

    A member is either a property or a method.
    """

    def __init__(
        self, instance: "Expression", name: Identifier, original_node: ast.AST
    ) -> None:
        """Initialize with the given values."""
        Expression.__init__(self, original_node=original_node)
        self.instance = instance
        self.name = name

    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        return transformer.transform_member(self)

    def visit(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        visitor.visit_member(self)


class Index(Expression):
    """Represent an index access on a collection."""

    def __init__(
        self, collection: "Expression", index: "Expression", original_node: ast.AST
    ) -> None:
        """Initialize with the given values."""
        Expression.__init__(self, original_node=original_node)
        self.collection = collection
        self.index = index

    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        return transformer.transform_index(self)

    def visit(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        visitor.visit_index(self)


class Comparator(enum.Enum):
    """List comparison operands."""

    LT = "LT"
    LE = "LE"
    GT = "GT"
    GE = "GE"
    EQ = "EQ"
    NE = "NE"


class Comparison(Expression):
    """Represent a comparison operation."""

    def __init__(
        self,
        left: "Expression",
        op: Comparator,
        right: "Expression",
        original_node: ast.AST,
    ) -> None:
        Expression.__init__(self, original_node=original_node)
        self.left = left
        self.op = op
        self.right = right

    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        return transformer.transform_comparison(self)

    def visit(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        visitor.visit_comparison(self)


class IsIn(Expression):
    """Represent the membership ``x in A``."""

    def __init__(
        self, member: "Expression", container: "Expression", original_node: ast.AST
    ) -> None:
        """Initialize with the given values."""
        Expression.__init__(self, original_node=original_node)
        self.member = member
        self.container = container

    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        return transformer.transform_is_in(self)

    def visit(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        visitor.visit_is_in(self)


class Implication(Expression):
    """Represent an implication of the form ``A => B``."""

    def __init__(
        self, antecedent: "Expression", consequent: "Expression", original_node: ast.AST
    ) -> None:
        """Initialize with the given values."""
        Expression.__init__(self, original_node=original_node)
        self.antecedent = antecedent
        self.consequent = consequent

    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        return transformer.transform_implication(self)

    def visit(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        visitor.visit_implication(self)


class MethodCall(Expression):
    """Represent a method call."""

    def __init__(
        self, member: Member, args: Sequence["Expression"], original_node: ast.AST
    ) -> None:
        """Initialize with the given values."""
        Expression.__init__(self, original_node=original_node)
        self.member = member
        self.args = args

    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        return transformer.transform_method_call(self)

    def visit(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        visitor.visit_method_call(self)


class Name(Expression):
    """Represent access to a variable with the given name."""

    def __init__(self, identifier: Identifier, original_node: ast.AST) -> None:
        """Initialize with the given values."""
        Expression.__init__(self, original_node=original_node)
        self.identifier = identifier

    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        return transformer.transform_name(self)

    def visit(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        visitor.visit_name(self)


class FunctionCall(Expression):
    """Represent a function call."""

    def __init__(
        self, name: Name, args: Sequence["Expression"], original_node: ast.AST
    ) -> None:
        """Initialize with the given values."""
        Expression.__init__(self, original_node=original_node)
        self.name = name
        self.args = args

    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        return transformer.transform_function_call(self)

    def visit(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        visitor.visit_function_call(self)


class Constant(Expression):
    """Represent a constant value."""

    def __init__(
        self, value: Union[bool, int, float, str], original_node: ast.AST
    ) -> None:
        """Initialize with the given values."""
        Expression.__init__(self, original_node=original_node)
        self.value = value

    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        return transformer.transform_constant(self)

    def visit(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        visitor.visit_constant(self)


class IsNone(Expression):
    """Represent a check whether something ``is None``."""

    def __init__(self, value: Expression, original_node: ast.AST) -> None:
        """Initialize with the given values."""
        Expression.__init__(self, original_node=original_node)
        self.value = value

    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        return transformer.transform_is_none(self)

    def visit(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        visitor.visit_is_none(self)


class IsNotNone(Expression):
    """Represent a check whether something ``is not None``."""

    def __init__(self, value: Expression, original_node: ast.AST) -> None:
        """Initialize with the given values."""
        Expression.__init__(self, original_node=original_node)
        self.value = value

    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        return transformer.transform_is_not_none(self)

    def visit(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        visitor.visit_is_not_none(self)


class Not(Expression):
    """Represent a negation."""

    def __init__(self, operand: Expression, original_node: ast.AST) -> None:
        """Initialize with the given values."""
        Expression.__init__(self, original_node=original_node)
        self.operand = operand

    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        return transformer.transform_not(self)

    def visit(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        visitor.visit_not(self)


class And(Expression):
    """Represent a conjunction."""

    def __init__(self, values: Sequence[Expression], original_node: ast.AST) -> None:
        Expression.__init__(self, original_node=original_node)
        self.values = values

    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        return transformer.transform_and(self)

    def visit(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        visitor.visit_and(self)


class Or(Expression):
    """Represent a disjunction."""

    def __init__(self, values: Sequence[Expression], original_node: ast.AST) -> None:
        Expression.__init__(self, original_node=original_node)
        self.values = values

    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        return transformer.transform_or(self)

    def visit(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        visitor.visit_or(self)


class Add(Expression):
    """Represent an addition."""

    def __init__(
        self, left: Expression, right: Expression, original_node: ast.AST
    ) -> None:
        Expression.__init__(self, original_node=original_node)
        self.left = left
        self.right = right

    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        return transformer.transform_add(self)

    def visit(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        visitor.visit_add(self)


class Sub(Expression):
    """Represent a subtraction."""

    def __init__(
        self, left: Expression, right: Expression, original_node: ast.AST
    ) -> None:
        Expression.__init__(self, original_node=original_node)
        self.left = left
        self.right = right

    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        return transformer.transform_sub(self)

    def visit(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        visitor.visit_sub(self)


class FormattedValue(Node):
    """Represent a formatted value in a :py:class`JoinedStr`."""

    def __init__(self, value: Expression, original_node: ast.AST) -> None:
        """Initialize with the given values."""
        Node.__init__(self, original_node=original_node)

        self.value = value

    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        return transformer.transform_formatted_value(self)

    def visit(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        visitor.visit_formatted_value(self)


class JoinedStr(Expression):
    """Represent a string interpolation."""

    def __init__(
        self, values: Sequence[Union[str, FormattedValue]], original_node: ast.AST
    ) -> None:
        """Initialize with the given values."""
        Expression.__init__(self, original_node=original_node)

        self.values = values

    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        return transformer.transform_joined_str(self)

    def visit(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        visitor.visit_joined_str(self)


class ForEach(Node):
    """Structure the information about the iteration over a collection."""

    def __init__(
        self, variable: Name, iteration: Expression, original_node: ast.AST
    ) -> None:
        """Initialize with the given values."""
        Node.__init__(self, original_node=original_node)
        self.variable = variable
        self.iteration = iteration

    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        return transformer.transform_for_each(self)

    def visit(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        visitor.visit_for_each(self)


class ForRange(Node):
    """Structure the information about the iteration over a range of integers."""

    def __init__(
        self, variable: Name, start: Expression, end: Expression, original_node: ast.AST
    ) -> None:
        """Initialize with the given values."""
        Node.__init__(self, original_node=original_node)
        self.variable = variable
        self.start = start
        self.end = end

    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        return transformer.transform_for_range(self)

    def visit(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        visitor.visit_for_range(self)


ForUnion = Union[ForEach, ForRange]


class Any(Expression):
    """Represent an ``any(...)`` expression."""

    def __init__(
        self, generator: ForUnion, condition: Expression, original_node: ast.AST
    ) -> None:
        """Initialize with the given values."""
        Expression.__init__(self, original_node=original_node)
        self.generator = generator
        self.condition = condition

    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        return transformer.transform_any(self)

    def visit(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        visitor.visit_any(self)


class All(Expression):
    """Represent an ``all(...)`` expression."""

    def __init__(
        self, generator: ForUnion, condition: Expression, original_node: ast.AST
    ) -> None:
        """Initialize with the given values."""
        Expression.__init__(self, original_node=original_node)
        self.generator = generator
        self.condition = condition

    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        return transformer.transform_all(self)

    def visit(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        visitor.visit_all(self)


class Assignment(Statement):
    """Represent an assignment of a single value to a single target."""

    def __init__(
        self, target: Expression, value: Expression, original_node: ast.AST
    ) -> None:
        """Initialize with the given values."""
        Statement.__init__(self, original_node=original_node)
        self.target = target
        self.value = value

    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        return transformer.transform_assignment(self)

    def visit(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        visitor.visit_assignment(self)


class Return(Statement):
    """Represent a return statement with a single return value."""

    def __init__(self, value: Optional[Expression], original_node: ast.AST) -> None:
        """Initialize with the given values."""
        Statement.__init__(self, original_node=original_node)
        self.value = value

    def transform(self, transformer: "Transformer[T]") -> T:
        """Accept the transformer."""
        return transformer.transform_return(self)

    def visit(self, visitor: "Visitor") -> None:
        """Accept the visitor."""
        visitor.visit_return(self)


class Visitor(DBC):
    """
    Visit all the nodes in the AST.

    The default action for the generic visitor is to descend recursively into all the
    children nodes.
    """

    def visit(self, node: Node) -> None:
        """Dispatch to the appropriate visitation method."""
        node.visit(self)

    def visit_member(self, node: Member) -> None:
        """Visit a member."""
        self.visit(node.instance)

    def visit_index(self, node: Index) -> None:
        """Visit an index access."""
        self.visit(node.collection)
        self.visit(node.index)

    def visit_comparison(self, node: Comparison) -> None:
        """Visit a comparison."""
        self.visit(node.left)
        self.visit(node.right)

    def visit_is_in(self, node: IsIn) -> None:
        """Visit a membership relation."""
        self.visit(node.member)
        self.visit(node.container)

    def visit_implication(self, node: Implication) -> None:
        """Visit an implication."""
        self.visit(node.antecedent)
        self.visit(node.consequent)

    def visit_method_call(self, node: MethodCall) -> None:
        """Visit a method call."""
        self.visit(node.member)
        for arg in node.args:
            self.visit(arg)

    def visit_function_call(self, node: FunctionCall) -> None:
        """Visit a function call."""
        for arg in node.args:
            self.visit(arg)

    def visit_constant(self, node: Constant) -> None:
        """Visit a constant."""
        pass

    def visit_is_none(self, node: IsNone) -> None:
        """Visit an is-none check."""
        self.visit(node.value)

    def visit_is_not_none(self, node: IsNotNone) -> None:
        """Visit an is-not-none check."""
        self.visit(node.value)

    def visit_name(self, node: Name) -> None:
        """Visit a variable access."""
        pass

    def visit_not(self, node: Not) -> None:
        """Visit a negation."""
        self.visit(node.operand)

    def visit_and(self, node: And) -> None:
        """Visit a conjunction."""
        for value in node.values:
            self.visit(value)

    def visit_or(self, node: Or) -> None:
        """Visit a disjunction."""
        for value in node.values:
            self.visit(value)

    def visit_add(self, node: Add) -> None:
        """Visit an addition."""
        self.visit(node.left)
        self.visit(node.right)

    def visit_sub(self, node: Sub) -> None:
        """Visit a subtraction."""
        self.visit(node.left)
        self.visit(node.right)

    def visit_formatted_value(self, node: FormattedValue) -> None:
        """Visit a formatted value in a joined string."""
        self.visit(node.value)

    def visit_joined_str(self, node: JoinedStr) -> None:
        """Visit a string interpolation."""
        for value in node.values:
            if isinstance(value, str):
                pass
            elif isinstance(value, FormattedValue):
                self.visit(value)
            else:
                assert_never(value)

    def visit_for_each(self, node: ForEach) -> None:
        """Visit a for-each in a generator."""
        self.visit(node.iteration)

    def visit_for_range(self, node: ForRange) -> None:
        """Visit a for-range in a generator."""
        self.visit(node.start)
        self.visit(node.end)

    def visit_any(self, node: Any) -> None:
        """Visit an ``any(...)`` expression."""
        self.visit(node.generator)
        self.visit(node.condition)

    def visit_all(self, node: All) -> None:
        """Visit an ``all(...)`` expression."""
        self.visit(node.generator)
        self.visit(node.condition)

    def visit_assignment(self, node: Assignment) -> None:
        """Visit an assignment statement."""
        self.visit(node.target)
        self.visit(node.value)

    def visit_return(self, node: Return) -> None:
        """Visit a return statement."""
        if node.value is not None:
            self.visit(node.value)


class Transformer(Generic[T], DBC):
    """Transform our AST into something."""

    def transform(self, node: Node) -> T:
        """Dispatch to the appropriate transformation method."""
        return node.transform(self)

    @abc.abstractmethod
    def transform_member(self, node: Member) -> T:
        """Transform a member to something."""
        raise NotImplementedError(f"{node=}")

    @abc.abstractmethod
    def transform_index(self, node: Index) -> T:
        """Transform an index access to something."""
        raise NotImplementedError(f"{node=}")

    @abc.abstractmethod
    def transform_comparison(self, node: Comparison) -> T:
        """Transform a comparison to something."""
        raise NotImplementedError(f"{node=}")

    @abc.abstractmethod
    def transform_is_in(self, node: IsIn) -> T:
        """Transform a membership relation to something."""
        raise NotImplementedError(f"{node=}")

    @abc.abstractmethod
    def transform_implication(self, node: Implication) -> T:
        """Transform an implication to something."""
        raise NotImplementedError(f"{node=}")

    @abc.abstractmethod
    def transform_method_call(self, node: MethodCall) -> T:
        """Transform a method call into something."""
        raise NotImplementedError(f"{node=}")

    @abc.abstractmethod
    def transform_function_call(self, node: FunctionCall) -> T:
        """Transform a function call into something."""
        raise NotImplementedError(f"{node=}")

    @abc.abstractmethod
    def transform_constant(self, node: Constant) -> T:
        """Transform a constant into something."""
        raise NotImplementedError(f"{node=}")

    @abc.abstractmethod
    def transform_is_none(self, node: IsNone) -> T:
        """Transform an is-none check into something."""
        raise NotImplementedError(f"{node=}")

    @abc.abstractmethod
    def transform_is_not_none(self, node: IsNotNone) -> T:
        """Transform an is-not-none check into something."""
        raise NotImplementedError(f"{node=}")

    @abc.abstractmethod
    def transform_name(self, node: Name) -> T:
        """Transform variable access into something."""
        raise NotImplementedError(f"{node=}")

    @abc.abstractmethod
    def transform_not(self, node: Not) -> T:
        """Transform a negation into something."""
        raise NotImplementedError(f"{node=}")

    @abc.abstractmethod
    def transform_and(self, node: And) -> T:
        """Transform a conjunction into something."""
        raise NotImplementedError(f"{node=}")

    @abc.abstractmethod
    def transform_or(self, node: Or) -> T:
        """Transform a disjunction into something."""
        raise NotImplementedError(f"{node=}")

    @abc.abstractmethod
    def transform_add(self, node: Add) -> T:
        """Transform an addition into something."""
        raise NotImplementedError(f"{node=}")

    @abc.abstractmethod
    def transform_sub(self, node: Sub) -> T:
        """Transform a subtraction into something."""
        raise NotImplementedError(f"{node=}")

    @abc.abstractmethod
    def transform_formatted_value(self, node: FormattedValue) -> T:
        """Transform a formatted value of a joined string into something."""
        raise NotImplementedError(f"{node=}")

    @abc.abstractmethod
    def transform_joined_str(self, node: JoinedStr) -> T:
        """Transform a string interpolation into something."""
        raise NotImplementedError(f"{node=}")

    @abc.abstractmethod
    def transform_for_each(self, node: ForEach) -> T:
        """Transform the for-each in a generator into something."""
        raise NotImplementedError(f"{node=}")

    @abc.abstractmethod
    def transform_for_range(self, node: ForRange) -> T:
        """Transform the for-range in a generator into something."""
        raise NotImplementedError(f"{node=}")

    @abc.abstractmethod
    def transform_any(self, node: Any) -> T:
        """Transform an ``any(...)`` into something."""
        raise NotImplementedError(f"{node=}")

    @abc.abstractmethod
    def transform_all(self, node: All) -> T:
        """Transform an ``all(...)`` into something."""
        raise NotImplementedError(f"{node=}")

    @abc.abstractmethod
    def transform_assignment(self, node: Assignment) -> T:
        """Transform an assignment into something."""
        raise NotImplementedError(f"{node=}")

    @abc.abstractmethod
    def transform_return(self, node: Return) -> T:
        """Transform a return statement into something."""
        raise NotImplementedError(f"{node=}")


class _StringifyTransformer(Transformer[stringify.Entity]):
    """Transform a node into a stringifiable representation."""

    def transform(self, node: Node) -> stringify.Entity:
        """Dispatch to the appropriate transformation method."""
        result = node.transform(self)
        stringify.assert_compares_against_dict(entity=result, obj=node)
        return result

    def transform_member(self, node: Member) -> stringify.Entity:
        return stringify.Entity(
            name=node.__class__.__name__,
            properties=[
                stringify.Property("instance", self.transform(node.instance)),
                stringify.Property("name", node.name),
                stringify.PropertyEllipsis("original_node", node.original_node),
            ],
        )

    def transform_index(self, node: Index) -> stringify.Entity:
        return stringify.Entity(
            name=node.__class__.__name__,
            properties=[
                stringify.Property("collection", self.transform(node.collection)),
                stringify.Property("index", self.transform(node.index)),
                stringify.PropertyEllipsis("original_node", node.original_node),
            ],
        )

    def transform_comparison(self, node: Comparison) -> stringify.Entity:
        return stringify.Entity(
            name=node.__class__.__name__,
            properties=[
                stringify.Property("left", self.transform(node.left)),
                stringify.Property("op", str(node.op.value)),
                stringify.Property("right", self.transform(node.right)),
                stringify.PropertyEllipsis("original_node", node.original_node),
            ],
        )

    def transform_is_in(self, node: IsIn) -> stringify.Entity:
        return stringify.Entity(
            name=node.__class__.__name__,
            properties=[
                stringify.Property("member", self.transform(node.member)),
                stringify.Property("container", self.transform(node.container)),
                stringify.PropertyEllipsis("original_node", node.original_node),
            ],
        )

    def transform_implication(self, node: Implication) -> stringify.Entity:
        return stringify.Entity(
            name=node.__class__.__name__,
            properties=[
                stringify.Property("antecedent", self.transform(node.antecedent)),
                stringify.Property("consequent", self.transform(node.consequent)),
                stringify.PropertyEllipsis("original_node", node.original_node),
            ],
        )

    def transform_method_call(self, node: MethodCall) -> stringify.Entity:
        return stringify.Entity(
            name=node.__class__.__name__,
            properties=[
                stringify.Property("member", self.transform(node.member)),
                stringify.Property("args", [self.transform(arg) for arg in node.args]),
                stringify.PropertyEllipsis("original_node", node.original_node),
            ],
        )

    def transform_function_call(self, node: FunctionCall) -> stringify.Entity:
        return stringify.Entity(
            name=node.__class__.__name__,
            properties=[
                stringify.Property("name", node.name.identifier),
                stringify.Property("args", [self.transform(arg) for arg in node.args]),
                stringify.PropertyEllipsis("original_node", node.original_node),
            ],
        )

    def transform_constant(self, node: Constant) -> stringify.Entity:
        return stringify.Entity(
            name=node.__class__.__name__,
            properties=[
                stringify.Property("value", node.value),
                stringify.PropertyEllipsis("original_node", node.original_node),
            ],
        )

    def transform_is_none(self, node: IsNone) -> stringify.Entity:
        return stringify.Entity(
            name=node.__class__.__name__,
            properties=[
                stringify.Property("value", self.transform(node.value)),
                stringify.PropertyEllipsis("original_node", node.original_node),
            ],
        )

    def transform_is_not_none(self, node: IsNotNone) -> stringify.Entity:
        return stringify.Entity(
            name=node.__class__.__name__,
            properties=[
                stringify.Property("value", self.transform(node.value)),
                stringify.PropertyEllipsis("original_node", node.original_node),
            ],
        )

    def transform_name(self, node: Name) -> stringify.Entity:
        return stringify.Entity(
            name=node.__class__.__name__,
            properties=[
                stringify.Property("identifier", node.identifier),
                stringify.PropertyEllipsis("original_node", node.original_node),
            ],
        )

    def transform_not(self, node: Not) -> stringify.Entity:
        return stringify.Entity(
            name=node.__class__.__name__,
            properties=[
                stringify.Property("operand", self.transform(node.operand)),
                stringify.PropertyEllipsis("original_node", node.original_node),
            ],
        )

    def transform_and(self, node: And) -> stringify.Entity:
        return stringify.Entity(
            name=node.__class__.__name__,
            properties=[
                stringify.Property(
                    "values", [self.transform(value) for value in node.values]
                ),
                stringify.PropertyEllipsis("original_node", node.original_node),
            ],
        )

    def transform_or(self, node: Or) -> stringify.Entity:
        return stringify.Entity(
            name=node.__class__.__name__,
            properties=[
                stringify.Property(
                    "values", [self.transform(value) for value in node.values]
                ),
                stringify.PropertyEllipsis("original_node", node.original_node),
            ],
        )

    def transform_add(self, node: Add) -> stringify.Entity:
        return stringify.Entity(
            name=node.__class__.__name__,
            properties=[
                stringify.Property("left", self.transform(node.left)),
                stringify.Property("right", self.transform(node.right)),
                stringify.PropertyEllipsis("original_node", node.original_node),
            ],
        )

    def transform_sub(self, node: Sub) -> stringify.Entity:
        return stringify.Entity(
            name=node.__class__.__name__,
            properties=[
                stringify.Property("left", self.transform(node.left)),
                stringify.Property("right", self.transform(node.right)),
                stringify.PropertyEllipsis("original_node", node.original_node),
            ],
        )

    def transform_formatted_value(self, node: FormattedValue) -> stringify.Entity:
        return stringify.Entity(
            name=node.__class__.__name__,
            properties=[
                stringify.Property("value", self.transform(node.value)),
                stringify.PropertyEllipsis("original_node", node.original_node),
            ],
        )

    def transform_joined_str(self, node: JoinedStr) -> stringify.Entity:
        values = []  # type: List[Union[stringify.Entity, str]]
        for value in node.values:
            if isinstance(value, str):
                values.append(value)
            elif isinstance(value, FormattedValue):
                values.append(self.transform(value))
            else:
                assert_never(value)

        return stringify.Entity(
            name=node.__class__.__name__,
            properties=[
                stringify.Property(
                    "values",
                    values,
                ),
                stringify.PropertyEllipsis("original_node", node.original_node),
            ],
        )

    def transform_for_each(self, node: ForEach) -> stringify.Entity:
        return stringify.Entity(
            name=node.__class__.__name__,
            properties=[
                stringify.Property("variable", self.transform(node.variable)),
                stringify.Property("iteration", self.transform(node.iteration)),
                stringify.PropertyEllipsis("original_node", node.original_node),
            ],
        )

    def transform_for_range(self, node: ForRange) -> stringify.Entity:
        return stringify.Entity(
            name=node.__class__.__name__,
            properties=[
                stringify.Property("variable", self.transform(node.variable)),
                stringify.Property("start", self.transform(node.start)),
                stringify.Property("end", self.transform(node.end)),
                stringify.PropertyEllipsis("original_node", node.original_node),
            ],
        )

    def transform_any(self, node: Any) -> stringify.Entity:
        return stringify.Entity(
            name=node.__class__.__name__,
            properties=[
                stringify.Property("generator", self.transform(node.generator)),
                stringify.Property("condition", self.transform(node.condition)),
                stringify.PropertyEllipsis("original_node", node.original_node),
            ],
        )

    def transform_all(self, node: All) -> stringify.Entity:
        return stringify.Entity(
            name=node.__class__.__name__,
            properties=[
                stringify.Property("generator", self.transform(node.generator)),
                stringify.Property("condition", self.transform(node.condition)),
                stringify.PropertyEllipsis("original_node", node.original_node),
            ],
        )

    def transform_assignment(self, node: Assignment) -> stringify.Entity:
        return stringify.Entity(
            name=node.__class__.__name__,
            properties=[
                stringify.Property("target", self.transform(node.target)),
                stringify.Property("value", self.transform(node.value)),
                stringify.PropertyEllipsis("original_node", node.original_node),
            ],
        )

    def transform_return(self, node: Return) -> stringify.Entity:
        return stringify.Entity(
            name=node.__class__.__name__,
            properties=[
                stringify.Property(
                    "value",
                    self.transform(node.value) if node.value is not None else None,
                ),
                stringify.PropertyEllipsis("original_node", node.original_node),
            ],
        )


def dump(node: Node) -> str:
    """Produce a string representation of the tree."""
    transformer = _StringifyTransformer()
    return stringify.dump(transformer.transform(node=node))


class RestrictedTransformer(Transformer[T]):
    """
    Transform our AST into something where only a part of the tree is handled.

    This class is helpful for cases where you are sure that certain nodes *can not*
    appear.
    """

    def transform_member(self, node: Member) -> T:
        """Transform a member to something."""
        raise AssertionError(f"Unexpected node: {dump(node)}")

    def transform_index(self, node: Index) -> T:
        """Transform an index access to something."""
        raise AssertionError(f"Unexpected node: {dump(node)}")

    def transform_comparison(self, node: Comparison) -> T:
        """Transform a comparison to something."""
        raise AssertionError(f"Unexpected node: {dump(node)}")

    def transform_is_in(self, node: IsIn) -> T:
        """Transform a membership relation to something."""
        raise NotImplementedError(f"{node=}")

    def transform_implication(self, node: Implication) -> T:
        """Transform an implication to something."""
        raise AssertionError(f"Unexpected node: {dump(node)}")

    def transform_method_call(self, node: MethodCall) -> T:
        """Transform a method call into something."""
        raise AssertionError(f"Unexpected node: {dump(node)}")

    def transform_function_call(self, node: FunctionCall) -> T:
        """Transform a function call into something."""
        raise AssertionError(f"Unexpected node: {dump(node)}")

    def transform_constant(self, node: Constant) -> T:
        """Transform a constant into something."""
        raise AssertionError(f"Unexpected node: {dump(node)}")

    def transform_is_none(self, node: IsNone) -> T:
        """Transform an is-none check into something."""
        raise AssertionError(f"Unexpected node: {dump(node)}")

    def transform_is_not_none(self, node: IsNotNone) -> T:
        """Transform an is-not-none check into something."""
        raise AssertionError(f"Unexpected node: {dump(node)}")

    def transform_name(self, node: Name) -> T:
        """Transform variable access into something."""
        raise AssertionError(f"Unexpected node: {dump(node)}")

    def transform_not(self, node: Not) -> T:
        """Transform a negation into something."""
        raise AssertionError(f"Unexpected node: {dump(node)}")

    def transform_and(self, node: And) -> T:
        """Transform a conjunction into something."""
        raise AssertionError(f"Unexpected node: {dump(node)}")

    def transform_or(self, node: Or) -> T:
        """Transform a disjunction into something."""
        raise AssertionError(f"Unexpected node: {dump(node)}")

    def transform_add(self, node: Add) -> T:
        """Transform an addition into something."""
        raise NotImplementedError(f"{node=}")

    def transform_sub(self, node: Sub) -> T:
        """Transform a subtraction into something."""
        raise NotImplementedError(f"{node=}")

    def transform_formatted_value(self, node: FormattedValue) -> T:
        """Transform a formatted value in a joined string into something."""
        raise AssertionError(f"Unexpected node: {dump(node)}")

    def transform_joined_str(self, node: JoinedStr) -> T:
        """Transform a string interpolation into something."""
        raise AssertionError(f"Unexpected node: {dump(node)}")

    def transform_for_each(self, node: ForEach) -> T:
        """Transform a for-each in a generator expression into something."""
        raise AssertionError(f"Unexpected node: {dump(node)}")

    def transform_for_range(self, node: ForRange) -> T:
        """Transform a for-range in a generator expression into something."""
        raise AssertionError(f"Unexpected node: {dump(node)}")

    def transform_any(self, node: Any) -> T:
        """Transform an ``any(...)`` expression into something."""
        raise AssertionError(f"Unexpected node: {dump(node)}")

    def transform_all(self, node: All) -> T:
        """Transform an ``all(...)`` expression into something."""
        raise AssertionError(f"Unexpected node: {dump(node)}")

    def transform_assignment(self, node: Assignment) -> T:
        """Transform an assignment into something."""
        raise AssertionError(f"Unexpected node: {dump(node)}")

    def transform_return(self, node: Return) -> T:
        """Transform a return statement into something."""
        raise AssertionError(f"Unexpected node: {dump(node)}")


class _IterationTransformer(Transformer[Iterator[Node]]):
    """Transform a node into an iterator over nodes."""

    def transform(self, node: Node) -> Iterator[Node]:
        """Dispatch to the appropriate transformation method."""
        yield from node.transform(self)

    def transform_member(self, node: Member) -> Iterator[Node]:
        yield node
        yield from self.transform(node.instance)

    def transform_index(self, node: Index) -> Iterator[Node]:
        yield node
        yield from self.transform(node.collection)
        yield from self.transform(node.index)

    def transform_comparison(self, node: Comparison) -> Iterator[Node]:
        yield node
        yield from self.transform(node.left)
        yield from self.transform(node.right)

    def transform_is_in(self, node: IsIn) -> Iterator[Node]:
        yield node
        yield from self.transform(node.member)
        yield from self.transform(node.container)

    def transform_implication(self, node: Implication) -> Iterator[Node]:
        yield node
        yield from self.transform(node.antecedent)
        yield from self.transform(node.consequent)

    def transform_method_call(self, node: MethodCall) -> Iterator[Node]:
        yield node
        yield from self.transform(node.member)
        for arg in node.args:
            yield from self.transform(arg)

    def transform_function_call(self, node: FunctionCall) -> Iterator[Node]:
        yield node
        for arg in node.args:
            yield from self.transform(arg)

    def transform_constant(self, node: Constant) -> Iterator[Node]:
        yield node

    def transform_is_none(self, node: IsNone) -> Iterator[Node]:
        yield node
        yield from self.transform(node.value)

    def transform_is_not_none(self, node: IsNotNone) -> Iterator[Node]:
        yield node
        yield from self.transform(node.value)

    def transform_name(self, node: Name) -> Iterator[Node]:
        yield node

    def transform_not(self, node: Not) -> Iterator[Node]:
        yield node
        yield from self.transform(node.operand)

    def transform_and(self, node: And) -> Iterator[Node]:
        yield node
        for value in node.values:
            yield from self.transform(value)

    def transform_or(self, node: Or) -> Iterator[Node]:
        yield node
        for value in node.values:
            yield from self.transform(value)

    def transform_add(self, node: Add) -> Iterator[Node]:
        yield node
        yield from self.transform(node.left)
        yield from self.transform(node.right)

    def transform_sub(self, node: Sub) -> Iterator[Node]:
        yield node
        yield from self.transform(node.left)
        yield from self.transform(node.right)

    def transform_formatted_value(self, node: FormattedValue) -> Iterator[Node]:
        yield node
        yield from self.transform(node.value)

    def transform_joined_str(self, node: JoinedStr) -> Iterator[Node]:
        yield node

        for value in node.values:
            if isinstance(value, str):
                pass
            elif isinstance(value, FormattedValue):
                yield from self.transform(value)
            else:
                assert_never(value)

    def transform_for_each(self, node: ForEach) -> Iterator[Node]:
        yield node
        yield from self.transform(node.variable)
        yield from self.transform(node.iteration)

    def transform_for_range(self, node: ForRange) -> Iterator[Node]:
        yield node
        yield from self.transform(node.variable)
        yield from self.transform(node.start)
        yield from self.transform(node.end)

    def transform_any(self, node: Any) -> Iterator[Node]:
        yield node
        yield from self.transform(node.generator)
        yield from self.transform(node.condition)

    def transform_all(self, node: All) -> Iterator[Node]:
        yield node
        yield from self.transform(node.generator)
        yield from self.transform(node.condition)

    def transform_assignment(self, node: Assignment) -> Iterator[Node]:
        yield node
        yield from self.transform(node.target)
        yield from self.transform(node.value)

    def transform_return(self, node: Return) -> Iterator[Node]:
        yield node
        if node.value is not None:
            yield from self.transform(node.value)


_ITERATION_TRANSFORMER = _IterationTransformer()


def over_nodes(node: Node) -> Iterator[Node]:
    """Iterate recursively over the ``node``, including the ``node`` itself."""
    yield from _ITERATION_TRANSFORMER.transform(node)



"""Generate the code to test JSON de/serialization of enumerations."""

import io
from typing import List

from icontract import ensure

from aas_core_codegen import intermediate
from aas_core_codegen.common import Stripped, Identifier, indent_but_first_line
from aas_core_codegen.python import common as python_common, naming as python_naming
from aas_core_codegen.python.common import (
    INDENT as I,
    INDENT2 as II,
)


def _generate_test_case(symbol_table: intermediate.SymbolTable) -> Stripped:
    test_methods = []  # type: List[Stripped]

    for enumeration in symbol_table.enumerations:
        test_method_name = python_naming.method_name(
            Identifier(f"test_{enumeration.name}")
        )

        from_jsonable = python_naming.function_name(
            Identifier(f"{enumeration.name}_from_jsonable")
        )

        literals_joined = ",\n".join(
            repr(literal.value) for literal in enumeration.literals
        )

        test_methods.append(
            Stripped(
                f"""\
def {test_method_name}(self) -> None:
{I}for jsonable in [
{II}{indent_but_first_line(literals_joined, II)}
{I}]:
{II}enum_literal = aas_jsonization.{from_jsonable}(jsonable)

{II}self.assertEqual(enum_literal.value, jsonable)"""
            ),
        )

    body = "\n\n".join(test_methods) if len(test_methods) > 0 else "pass"

    return Stripped(
        f"""\
class TestRoundTrips(unittest.TestCase):
{I}{indent_but_first_line(body, I)}"""
    )


@ensure(
    lambda result: result.endswith("\n"),
    "Trailing newline mandatory for valid end-of-files",
)
def generate(
    symbol_table: intermediate.SymbolTable,
    qualified_module_name: python_common.QualifiedModuleName,
) -> str:
    """
    Generate the code to test JSON de/serialization of enumerations.

    The ``qualified_module_name`` indicates the fully-qualified name of the base module.
    """
    blocks = [
        Stripped(
            '''\
"""Test JSON de/serialization of enumerations."""'''
        ),
        python_common.WARNING,
        Stripped(
            """\
# pylint: disable=missing-docstring"""
        ),
        Stripped(
            """\
import unittest"""
        ),
        Stripped(
            f"""\
import {qualified_module_name}.jsonization as aas_jsonization"""
        ),
        _generate_test_case(symbol_table=symbol_table),
        Stripped(
            f"""\
if __name__ == "__main__":
{I}unittest.main()"""
        ),
        python_common.WARNING,
    ]

    out = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            out.write("\n\n\n")

        out.write(block)

    out.write("\n")

    return out.getvalue()



"""Generate C++ visitors to iterate over instances."""

import io
from typing import (
    List,
    Tuple,
)

from icontract import ensure

from aas_core_codegen import intermediate
from aas_core_codegen.common import (
    Identifier,
    assert_never,
    Stripped,
    indent_but_first_line,
)
from aas_core_codegen.cpp import (
    common as cpp_common,
    naming as cpp_naming,
)
from aas_core_codegen.cpp.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
)


# region Generation


def _generate_mutating_visitor_interface(
    symbol_table: intermediate.SymbolTable,
) -> Stripped:
    """Generate the interface for a mutating visitor."""
    visit_methods = []  # type: List[Stripped]

    for cls in symbol_table.concrete_classes:
        method_name = cpp_naming.method_name(Identifier(f"visit_{cls.name}"))
        that_type = cpp_naming.interface_name(cls.name)

        visit_methods.append(
            Stripped(
                f"""\
virtual void {method_name}(
{I}const std::shared_ptr<types::{that_type}>& that
) = 0;"""
            )
        )

    visit_methods_joined = "\n".join(visit_methods)

    return Stripped(
        f"""\
/**
 * Provide an interface for a recursive mutating visitor on an instance.
 */
class IVisitor {{
 public:
{I}/**
{I} * Visit \\p that instance and recursively visit all the instances
{I} * referenced from \\p that instance.
{I} *
{I} * We use const references to shared pointers here for efficiency in case you want,
{I} * say, to share ownership over instances in your own external containers. Since
{I} * we do not make copies of the shared pointers, it is very important that
{I} * the given shared pointers outlive the visitor, lest cause undefined behavior.
{I} * See these StackOverflow questions:
{I} * * https://stackoverflow.com/questions/12002480/passing-stdshared-ptr-to-constructors/12002668#12002668
{I} * * https://stackoverflow.com/questions/3310737/should-we-pass-a-shared-ptr-by-reference-or-by-value
{I} * * https://stackoverflow.com/questions/37610494/passing-const-shared-ptrt-versus-just-shared-ptrt-as-parameter
{I} *
{I} * Changing the references during the visitation results in undefined
{I} * behavior. This follows how STL deals with modifications to containers, see:
{I} * https://stackoverflow.com/questions/6438086/iterator-invalidation-rules-for-c-containers
{I} *
{I} * \\param that instance to be visited recursively
{I} */
{I}virtual void Visit(const std::shared_ptr<types::IClass>& that) = 0;
{I}virtual ~IVisitor() = default;

 protected:
{I}{indent_but_first_line(visit_methods_joined, I)}
}};  // class IVisitor"""
    )


def _generate_mutating_abstract_visitor_definition() -> Stripped:
    """Generate the definition of an abstract mutating visitor."""
    return Stripped(
        f"""\
/**
 * Provide an abstract recursive mutating visitor on an instance.
 */
class AbstractVisitor
{II}: public IVisitor {{
 public:
{I}void Visit(const std::shared_ptr<types::IClass>& that) override;
{I}~AbstractVisitor() override = default;
}};  // class AbstractVisitor"""
    )


def _generate_mutating_pass_through_visitor_definition(
    symbol_table: intermediate.SymbolTable,
) -> Stripped:
    """Generate the definition of a no-op mutating visitor."""
    visit_methods = []  # type: List[Stripped]

    for cls in symbol_table.concrete_classes:
        method_name = cpp_naming.method_name(Identifier(f"visit_{cls.name}"))
        that_type = cpp_naming.interface_name(cls.name)
        visit_methods.append(
            Stripped(
                f"""\
void {method_name}(
{I}const std::shared_ptr<types::{that_type}>& that
) override;"""
            )
        )

    visit_methods_joined = "\n".join(visit_methods)

    return Stripped(
        f"""\
/**
 * \\brief Provide a mutating, recursive and no-op visitor on an instance.
 *
 * Usually, you want to inherit from this visitor and override one or more of its
 * visitation methods.
 */
class PassThroughVisitor
{II}: public AbstractVisitor {{
 public:
{I}~PassThroughVisitor() override = default;

 protected:
{I}{indent_but_first_line(visit_methods_joined, I)}
}};  // class PassThroughVisitor"""
    )


# fmt: off
@ensure(
    lambda result:
    result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate_header(
    symbol_table: intermediate.SymbolTable, library_namespace: Stripped
) -> str:
    """Generate the C++ header code of the visitors."""
    namespace = Stripped(f"{library_namespace}::visitation")

    include_guard_var = cpp_common.include_guard_var(namespace)

    include_prefix_path = cpp_common.generate_include_prefix_path(library_namespace)

    blocks = [
        Stripped(
            f"""\
#ifndef {include_guard_var}
#define {include_guard_var}"""
        ),
        cpp_common.WARNING,
        Stripped(
            f'''\
#include "{include_prefix_path}/types.hpp"'''
        ),
        cpp_common.generate_namespace_opening(library_namespace),
        Stripped(
            """\
/**
 * \\defgroup visitation Iterate and modify instances through visitors.
 * @{
 */
namespace visitation {"""
        ),
        _generate_mutating_visitor_interface(symbol_table=symbol_table),
        _generate_mutating_abstract_visitor_definition(),
        _generate_mutating_pass_through_visitor_definition(symbol_table=symbol_table),
    ]  # type: List[Stripped]

    blocks.extend(
        [
            Stripped(
                """\
}  // namespace visitation
/**@*/"""
            ),
            cpp_common.generate_namespace_closing(library_namespace),
            cpp_common.WARNING,
            Stripped(f"#endif  // {include_guard_var}"),
        ]
    )

    out = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            out.write("\n\n")

        out.write(block)

    out.write("\n")

    return out.getvalue()


def _generate_dispatching_visit_switch_statement(
    symbol_table: intermediate.SymbolTable,
) -> Stripped:
    """Generate the dispatching statement for the visitor's main ``Visit`` method."""
    case_blocks = []  # type: List[Stripped]
    for cls in symbol_table.concrete_classes:
        model_type_literal = cpp_naming.enum_literal_name(cls.name)
        visit_name = cpp_naming.method_name(Identifier(f"visit_{cls.name}"))
        interface_name = cpp_naming.interface_name(cls.name)

        case_blocks.append(
            Stripped(
                f"""\
case types::ModelType::{model_type_literal}:
{I}{visit_name}(
{II}std::dynamic_pointer_cast<
{III}types::{interface_name}
{II}>(that)
{I});
{I}break;"""
            )
        )

    case_blocks.append(
        Stripped(
            f"""\
default:
{I}throw std::logic_error(
{II}common::Concat(
{III}"Unexpected model type: ",
{III}std::to_string(
{IIII}static_cast<std::uint32_t>(that->model_type())
{III})
{II})
{I});"""
        )
    )

    case_blocks_joined = "\n".join(case_blocks)

    switch_stmt = Stripped(
        f"""\
// NOTE (mristin):
// We have to dynamically cast the pointers due to the virtual multiple
// inheritance, and also because we used shared pointers for references.
// If we used constant references instead of shared pointers, we could use
// a pattern such as double dispatch. However, this has the limitation that
// it would prevent us from collecting the instances in the visitor, such that
// they outlive the original structures.

switch (that->model_type()) {{
{I}{indent_but_first_line(case_blocks_joined, I)}
}}"""
    )

    return switch_stmt


def _generate_mutating_abstract_visitor_implementation(
    symbol_table: intermediate.SymbolTable,
) -> List[Stripped]:
    """Generate C++ implementation of the visitor's method(s)."""
    switch_stmt = _generate_dispatching_visit_switch_statement(
        symbol_table=symbol_table
    )

    blocks = [
        Stripped("// region AbstractVisitor"),
        Stripped(
            f"""\
void AbstractVisitor::Visit(
{I}const std::shared_ptr<types::IClass>& that
) {{
{I}{indent_but_first_line(switch_stmt, I)}
}}"""
        ),
        Stripped("// endregion"),
    ]  # type: List[Stripped]

    return blocks


def _generate_recursive_visit_for_property(
    prop: intermediate.Property, mutating: bool
) -> Stripped:
    """
    Generate the snippet to visit a property recursively.

    Empty result means the property can not be descended into.

    If ``mutating`` is set, use the mutable getter.
    """
    type_anno = intermediate.beneath_optional(prop.type_annotation)

    getter = (
        cpp_naming.mutable_getter_name(prop.name)
        if mutating
        else cpp_naming.getter_name(prop.name)
    )

    if isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
        maybe_var = cpp_naming.variable_name(Identifier(f"maybe_{prop.name}"))
        get_expr = Stripped(f"{maybe_var}.value()")
    else:
        get_expr = Stripped(f"that->{getter}()")

    code: Stripped

    if isinstance(type_anno, intermediate.PrimitiveTypeAnnotation):
        # No visits to primitive values.
        return Stripped("")

    elif isinstance(type_anno, intermediate.OurTypeAnnotation):
        if isinstance(type_anno.our_type, intermediate.Enumeration):
            # No visits to enumerations.
            return Stripped("")

        elif isinstance(type_anno.our_type, intermediate.ConstrainedPrimitive):
            # No visits to primitive values.
            return Stripped("")

        elif isinstance(
            type_anno.our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
        ):
            code = Stripped(
                f"""\
Visit(
{I}{get_expr}
);"""
            )
        else:
            assert_never(type_anno.our_type)
    elif isinstance(type_anno, intermediate.ListTypeAnnotation):
        assert isinstance(
            type_anno.items, intermediate.OurTypeAnnotation
        ) and isinstance(
            type_anno.items.our_type,
            (intermediate.AbstractClass, intermediate.ConcreteClass),
        ), (
            f"NOTE (mristin, 2023-03-29): We expect only lists of classes "
            f"at the moment, but you specified {type_anno}. "
            f"Please contact the developers if you need this feature."
        )

        item_type = cpp_common.generate_type_with_const_ref_if_applicable(
            type_annotation=type_anno.items,
            types_namespace=cpp_common.TYPES_NAMESPACE,
        )

        code = Stripped(
            f"""\
for (
{I}{indent_but_first_line(item_type, I)} item :
{I}{indent_but_first_line(get_expr, I)}
) {{
{I}Visit(item);
}}"""
        )

    else:
        assert_never(type_anno)

    if not isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
        return Stripped(
            f"""\
// {getter}
{code}"""
        )

    maybe_var = cpp_naming.variable_name(Identifier(f"maybe_{prop.name}"))
    maybe_type = cpp_common.generate_type_with_const_ref_if_applicable(
        type_annotation=prop.type_annotation, types_namespace=cpp_common.TYPES_NAMESPACE
    )

    return Stripped(
        f"""\
// region {getter}
{maybe_type} {maybe_var}(
{I}that->{getter}()
);
if ({maybe_var}.has_value()) {{
{I}{indent_but_first_line(code, I)}
}}
// endregion"""
    )


def _generate_pass_through_visit_body_for_class(
    cls: intermediate.ConcreteClass, mutating: bool
) -> Tuple[Stripped, bool]:
    """
    Generate the body of a pass-through visit member.

    If ``mutating`` is set, the pass-through is mutating.

    Return (code, True if any of the properties will be recursively visited)
    """
    blocks = []  # type: List[Stripped]

    for prop in cls.properties:
        code = _generate_recursive_visit_for_property(prop=prop, mutating=mutating)
        if code != "":
            blocks.append(code)

    if len(blocks) == 0:
        return Stripped("// No properties to be passed through."), False

    return Stripped("\n\n".join(blocks)), True


def _generate_mutating_pass_through_visitor_implementation(
    symbol_table: intermediate.SymbolTable,
) -> List[Stripped]:
    """Generate C++ implementation of the visitor's method(s)."""
    blocks = [
        Stripped("// region PassThroughVisitor"),
    ]

    for cls in symbol_table.concrete_classes:
        body, is_recursive = _generate_pass_through_visit_body_for_class(
            cls=cls, mutating=True
        )
        method_name = cpp_naming.method_name(Identifier(f"visit_{cls.name}"))
        that_type = cpp_naming.interface_name(cls.name)

        if is_recursive:
            blocks.append(
                Stripped(
                    f"""\
void PassThroughVisitor::{method_name}(
{I}const std::shared_ptr<types::{that_type}>& that
) {{
{I}{indent_but_first_line(body, I)}
}}"""
                )
            )
        else:
            # NOTE (mristin, 2024-01-12):
            # We need to signal to the compiler that this function does nothing, and
            # that we will not use the argument. Note that ``that`` is only used if
            # there is further recursion, see above in the if-body.
            blocks.append(
                Stripped(
                    f"""\
void PassThroughVisitor::{method_name}(
{I}const std::shared_ptr<types::{that_type}>&
) {{
{I}{indent_but_first_line(body, I)}
}}"""
                )
            )

    blocks.append(Stripped("// endregion"))

    return blocks


# fmt: off
@ensure(
    lambda result:
    result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate_implementation(
    symbol_table: intermediate.SymbolTable, library_namespace: Stripped
) -> str:
    """Generate the C++ implementation code of the visitors."""
    namespace = Stripped(f"{library_namespace}::visitation")

    include_prefix_path = cpp_common.generate_include_prefix_path(library_namespace)

    blocks = [
        cpp_common.WARNING,
        Stripped(
            f"""\
#include "{include_prefix_path}/types.hpp"
#include "{include_prefix_path}/stringification.hpp"
#include "{include_prefix_path}/visitation.hpp"

#pragma warning(push, 0)
#include <sstream>
#pragma warning(pop)"""
        ),
        cpp_common.generate_namespace_opening(namespace),
        *_generate_mutating_abstract_visitor_implementation(symbol_table=symbol_table),
        *_generate_mutating_pass_through_visitor_implementation(
            symbol_table=symbol_table
        ),
    ]  # type: List[Stripped]

    blocks.extend(
        [
            cpp_common.generate_namespace_closing(namespace),
            cpp_common.WARNING,
        ]
    )

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


# endregion



"""Generate Java code to handle asset administration shells based on the meta-model."""
from typing import TextIO

from aas_core_codegen import specific_implementations, run
from aas_core_codegen.java import (
    common as java_common,
    constants as java_constants,
    copying as java_copying,
    enhancing as java_enhancing,
    generation as java_generation,
    jsonization as java_jsonization,
    reporting as java_reporting,
    stringification as java_stringification,
    structure as java_structure,
    verification as java_verification,
    visitation as java_visitation,
    xmlization as java_xmlization,
)


def execute(context: run.Context, stdout: TextIO, stderr: TextIO) -> int:
    """Generate the code."""

    verified_ir_table, errors = java_structure.verify(symbol_table=context.symbol_table)

    if errors is not None:
        run.write_error_report(
            message=f"Failed to verify the intermediate symbol table "
            f"for generation of Java code"
            f"based on {context.model_path}",
            errors=[context.lineno_columner.error_message(error) for error in errors],
            stderr=stderr,
        )
        return 1

    assert verified_ir_table is not None

    _ = context
    _ = stderr

    package_key = specific_implementations.ImplementationKey("package.txt")
    package_text = context.spec_impls.get(package_key, None)
    if package_text is None:
        stderr.write(f"The namespace snippet is missing: {package_key}\n")
        return 1

    if not java_common.PACKAGE_IDENTIFIER_RE.fullmatch(package_text):
        stderr.write(
            f"The text from the snippet {package_key} "
            f"is not a valid namespace identifier: {package_text!r}\n"
        )
        return 1

    package = java_common.PackageIdentifier(package_text)

    # region Structure

    source_files, errors = java_structure.generate(
        symbol_table=verified_ir_table,
        package=package,
        spec_impls=context.spec_impls,
    )

    if errors is not None:
        run.write_error_report(
            message=f"Failed to generate the structures in the Java code "
            f"based on {context.model_path}",
            errors=[context.lineno_columner.error_message(error) for error in errors],
            stderr=stderr,
        )
        return 1

    assert source_files is not None

    (context.output_dir / "types" / java_common.CLASS_PKG).mkdir(
        exist_ok=True, parents=True
    )
    (context.output_dir / "types" / java_common.INTERFACE_PKG).mkdir(
        exist_ok=True, parents=True
    )
    (context.output_dir / "types" / java_common.ENUM_PKG).mkdir(
        exist_ok=True, parents=True
    )

    for source_file in source_files:
        pth = context.output_dir / "types" / source_file.name
        try:
            pth.write_text(source_file.content, encoding="utf-8")
        except Exception as exception:
            run.write_error_report(
                message=f"Failed to write the Java structures to {pth}",
                errors=[str(exception)],
                stderr=stderr,
            )
            return 1

    # endregion

    # region Visitation

    source_files, errors = java_visitation.generate(
        symbol_table=context.symbol_table, package=package
    )

    if errors is not None:
        run.write_error_report(
            message=f"Failed to generate the Java code for visitation "
            f"based on {context.model_path}",
            errors=[context.lineno_columner.error_message(error) for error in errors],
            stderr=stderr,
        )
        return 1

    assert source_files is not None

    (context.output_dir / "visitation").mkdir(exist_ok=True, parents=True)

    for source_file in source_files:
        pth = context.output_dir / "visitation" / source_file.name
        try:
            pth.write_text(source_file.content, encoding="utf-8")
        except Exception as exception:
            run.write_error_report(
                message=f"Failed to write the visitation Java code to {pth}",
                errors=[str(exception)],
                stderr=stderr,
            )
            return 1

    # endregion

    # region Constants

    code, errors = java_constants.generate(
        symbol_table=context.symbol_table,
        package=package,
    )

    if errors is not None:
        run.write_error_report(
            message=f"Failed to generate the constants in the Java code "
            f"based on {context.model_path}",
            errors=[context.lineno_columner.error_message(error) for error in errors],
            stderr=stderr,
        )
        return 1

    assert code is not None

    pth = context.output_dir / "constants" / "Constants.java"
    pth.parent.mkdir(exist_ok=True, parents=True)

    try:
        pth.write_text(code, encoding="utf-8")
    except Exception as exception:
        run.write_error_report(
            message=f"Failed to write the constants in the Java code to {pth}",
            errors=[str(exception)],
            stderr=stderr,
        )
        return 1

    # endregion

    # region Verification

    verify_errors = java_verification.verify(
        spec_impls=context.spec_impls,
        verification_functions=verified_ir_table.verification_functions,
    )

    if verify_errors is not None:
        run.write_error_report(
            message="Failed to verify for generation of Java verification code",
            errors=verify_errors,
            stderr=stderr,
        )
        return 1

    code, errors = java_verification.generate(
        symbol_table=verified_ir_table,
        package=package,
        spec_impls=context.spec_impls,
    )

    if errors is not None:
        run.write_error_report(
            message=f"Failed to generate the verification Java code "
            f"based on {context.model_path}",
            errors=[context.lineno_columner.error_message(error) for error in errors],
            stderr=stderr,
        )
        return 1

    assert code is not None

    pth = context.output_dir / "verification" / "Verification.java"
    pth.parent.mkdir(exist_ok=True, parents=True)

    try:
        pth.write_text(code, encoding="utf-8")
    except Exception as exception:
        run.write_error_report(
            message=f"Failed to write the verification Java code to {pth}",
            errors=[str(exception)],
            stderr=stderr,
        )
        return 1

    # endregion

    # region Reporting

    code = java_reporting.generate(package=package)

    pth = context.output_dir / "reporting" / "Reporting.java"
    pth.parent.mkdir(exist_ok=True, parents=True)

    try:
        pth.write_text(code, encoding="utf-8")
    except Exception as exception:
        run.write_error_report(
            message=f"Failed to write the reporting Java code to {pth}",
            errors=[str(exception)],
            stderr=stderr,
        )
        return 1

    # endregion

    # region Stringification

    code, errors = java_stringification.generate(
        symbol_table=context.symbol_table, package=package
    )

    if errors is not None:
        run.write_error_report(
            message=f"Failed to generate the stringification Java code "
            f"based on {context.model_path}",
            errors=[context.lineno_columner.error_message(error) for error in errors],
            stderr=stderr,
        )
        return 1

    assert code is not None

    pth = context.output_dir / "stringification" / "Stringification.java"
    pth.parent.mkdir(exist_ok=True)

    try:
        pth.write_text(code, encoding="utf-8")
    except Exception as exception:
        run.write_error_report(
            message=f"Failed to write the stringification Java code to {pth}",
            errors=[str(exception)],
            stderr=stderr,
        )
        return 1

    # endregion

    # region Xmlization

    code, errors = java_xmlization.generate(
        symbol_table=context.symbol_table,
        package=package,
        spec_impls=context.spec_impls,
    )

    if errors is not None:
        run.write_error_report(
            message=f"Failed to generate the xmlization Java code "
            f"based on {context.model_path}",
            errors=[context.lineno_columner.error_message(error) for error in errors],
            stderr=stderr,
        )
        return 1

    assert code is not None

    pth = context.output_dir / "xmlization" / "Xmlization.java"
    pth.parent.mkdir(exist_ok=True)

    try:
        pth.write_text(code, encoding="utf-8")
    except Exception as exception:
        run.write_error_report(
            message=f"Failed to write the xmlization Java code to {pth}",
            errors=[str(exception)],
            stderr=stderr,
        )
        return 1

    # endregion

    # region Jsonization

    code, errors = java_jsonization.generate(
        symbol_table=context.symbol_table,
        package=package,
        spec_impls=context.spec_impls,
    )

    if errors is not None:
        run.write_error_report(
            message=f"Failed to generate the jsonization Java code "
            f"based on {context.model_path}",
            errors=[context.lineno_columner.error_message(error) for error in errors],
            stderr=stderr,
        )
        return 1

    assert code is not None

    pth = context.output_dir / "jsonization" / "Jsonization.java"
    pth.parent.mkdir(exist_ok=True)

    try:
        pth.write_text(code, encoding="utf-8")
    except Exception as exception:
        run.write_error_report(
            message=f"Failed to write the jsonization Java code to {pth}",
            errors=[str(exception)],
            stderr=stderr,
        )
        return 1

    # endregion

    # region Copying

    code, errors = java_copying.generate(
        symbol_table=context.symbol_table,
        package=package,
        spec_impls=context.spec_impls,
    )

    if errors is not None:
        run.write_error_report(
            message=f"Failed to generate the Java code for shallow and deep copying "
            f"based on {context.model_path}",
            errors=[context.lineno_columner.error_message(error) for error in errors],
            stderr=stderr,
        )
        return 1

    assert code is not None

    pth = context.output_dir / "copying" / "Copying.java"
    pth.parent.mkdir(exist_ok=True, parents=True)

    try:
        pth.write_text(code, encoding="utf-8")
    except Exception as exception:
        run.write_error_report(
            message=f"Failed to write the Java code for shallow and deep copying "
            f"to {pth}",
            errors=[str(exception)],
            stderr=stderr,
        )
        return 1

    # endregion

    # region Enhancing

    source_files, errors = java_enhancing.generate(
        symbol_table=context.symbol_table,
        package=package,
        spec_impls=context.spec_impls,
    )

    if errors is not None:
        run.write_error_report(
            message=f"Failed to generate the Java code for enhancing "
            f"based on {context.model_path}",
            errors=[context.lineno_columner.error_message(error) for error in errors],
            stderr=stderr,
        )
        return 1

    assert source_files is not None

    (context.output_dir / "enhancing").mkdir(exist_ok=True, parents=True)

    for source_file in source_files:
        pth = context.output_dir / "enhancing" / source_file.name

        try:
            pth.write_text(source_file.content, encoding="utf-8")
        except Exception as exception:
            run.write_error_report(
                message=f"Failed to write the Java code for enhancing to {pth}",
                errors=[str(exception)],
                stderr=stderr,
            )
            return 1

    # endregion

    # region Generation

    source_files, errors = java_generation.generate(
        symbol_table=context.symbol_table,
        package=package,
    )

    if errors is not None:
        run.write_error_report(
            message=f"Failed to generate the Java code for generation "
            f"based on {context.model_path}",
            errors=[context.lineno_columner.error_message(error) for error in errors],
            stderr=stderr,
        )
        return 1

    assert source_files is not None

    (context.output_dir / "generation").mkdir(exist_ok=True, parents=True)

    for source_file in source_files:
        pth = context.output_dir / "generation" / source_file.name

        try:
            pth.write_text(source_file.content, encoding="utf-8")
        except Exception as exception:
            run.write_error_report(
                message=f"Failed to write the Java code for generation to {pth}",
                errors=[str(exception)],
                stderr=stderr,
            )
            return 1

    # endregion

    stdout.write(f"Code generated to: {context.output_dir}\n")
    return 0



"""Generate the test code for the verification of enums."""

from typing import List

from icontract import ensure

from aas_core_codegen import intermediate
from aas_core_codegen.common import Stripped, indent_but_first_line
from aas_core_codegen.csharp import common as csharp_common, naming as csharp_naming
from aas_core_codegen.csharp.common import (
    INDENT as I,
    INDENT2 as II,
)


# fmt: off
@ensure(
    lambda result: result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(
    namespace: csharp_common.NamespaceIdentifier,
    symbol_table: intermediate.SymbolTable,
) -> str:
    """
    Generate the test code for the verification of enums.

    The ``namespace`` indicates the fully-qualified name of the base project.
    """
    blocks = []  # type: List[Stripped]

    for enumeration in symbol_table.enumerations:
        enum_name = csharp_naming.enum_name(enumeration.name)

        assert (
            len(enumeration.literals) > 0
        ), f"Unexpected enumeration without literals: {enumeration.name}"

        literal_name = csharp_naming.enum_literal_name(enumeration.literals[0].name)

        blocks.append(
            Stripped(
                f"""\
[Test]
public void Test_{enum_name}_valid()
{{
{I}var errors = Aas.Verification.Verify{enum_name}(
{II}Aas.{enum_name}.{literal_name}).ToList();

{I}Assert.IsEmpty(errors);
}}  // void Test_{enum_name}_valid"""
            )
        )

        blocks.append(
            Stripped(
                f"""\
[Test]
public void Test_{enum_name}_invalid()
{{
{I}int valueAsInt = -1;
{I}Aas.{enum_name} value = (Aas.{enum_name})valueAsInt;

{I}var errors = Aas.Verification.Verify{enum_name}(
{II}value).ToList();

{I}Assert.AreEqual(1, errors.Count);
{I}Assert.AreEqual("Invalid {enum_name}: -1", errors[0].Cause);
}}  // void Test_{enum_name}_invalid"""
            )
        )

    blocks_joined = "\n\n".join(blocks)

    return f"""\
{csharp_common.WARNING}

using Aas = {namespace};  // renamed

using System.Linq;  // can't alias
using NUnit.Framework;  // can't alias

namespace {namespace}.Tests
{{
{I}public class TestVerificationOfEnums
{I}{{
{II}{indent_but_first_line(blocks_joined, II)}
{I}}}  // class TestVerificationOfEnums
}}  // namespace {namespace}.Tests

{csharp_common.WARNING}
"""


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""
Generate C++ identifiers based on the identifiers from the meta-model.

The methods all generate public names, unless their prefix indicates
otherwise.

We follow the Google C++ code style,
see: https://google.github.io/styleguide/cppguide.html#Naming
"""

from aas_core_codegen import naming
from aas_core_codegen.common import Identifier


def interface_name(identifier: Identifier) -> Identifier:
    """Generate a C++ interface name based on its meta-model ``identifier``."""
    return Identifier(f"I{naming.capitalized_camel_case(identifier)}")


def enum_name(identifier: Identifier) -> Identifier:
    """
    Generate a C++ name for an enum based on its meta-model ``identifier``.

    >>> enum_name(Identifier("something"))
    'Something'

    >>> enum_name(Identifier("URL_to_something"))
    'UrlToSomething'
    """
    return naming.capitalized_camel_case(identifier)


def enum_literal_name(literal_name: Identifier) -> Identifier:
    """
    Generate a C++ name for an enum literal.

    >>> enum_literal_name(Identifier("ID_short"))
    'kIdShort'
    """
    return Identifier(f"k{naming.capitalized_camel_case(literal_name)}")


def class_name(identifier: Identifier) -> Identifier:
    """
    Generate a C++ name for a class based on its meta-model ``identifier``.

    >>> class_name(Identifier("something"))
    'Something'

    >>> class_name(Identifier("URL_to_something"))
    'UrlToSomething'
    """
    return naming.capitalized_camel_case(identifier)


def getter_name(identifier: Identifier) -> Identifier:
    """
    Generate a C++ name for a property getter based on its meta-model ``identifier``.

    >>> getter_name(Identifier("something"))
    'something'

    >>> getter_name(Identifier("something_to_URL"))
    'something_to_url'
    """
    return naming.lower_snake_case(identifier)


def mutable_getter_name(identifier: Identifier) -> Identifier:
    """
    Generate a C++ name for a property getter based on its meta-model ``identifier``.

    >>> mutable_getter_name(Identifier("something"))
    'mutable_something'

    >>> mutable_getter_name(Identifier("something_to_URL"))
    'mutable_something_to_url'
    """
    return naming.lower_snake_case(Identifier(f"mutable_{identifier}"))


def setter_name(identifier: Identifier) -> Identifier:
    """
    Generate a C++ name for a property setter based on its meta-model ``identifier``.

    >>> setter_name(Identifier("something"))
    'set_something'

    >>> setter_name(Identifier("something_to_URL"))
    'set_something_to_url'
    """
    return naming.lower_snake_case(Identifier(f"set_{identifier}"))


def private_property_name(identifier: Identifier) -> Identifier:
    """
    Generate a C++ name for a private property based on the ``identifier``.

    >>> private_property_name(Identifier("something"))
    'something_'

    >>> private_property_name(Identifier("something_to_URL"))
    'something_to_url_'
    """
    return Identifier(f"{naming.lower_snake_case(identifier)}_")


def method_name(identifier: Identifier) -> Identifier:
    """
    Generate a C++ name for a member method based on its meta-model ``identifier``.

    >>> method_name(Identifier("do_something"))
    'DoSomething'

    >>> method_name(Identifier("do_something_to_URL"))
    'DoSomethingToUrl'
    """
    return naming.capitalized_camel_case(identifier)


def function_name(identifier: Identifier) -> Identifier:
    """
    Generate a name for a function from its meta-model ``identifier``.

    >>> function_name(Identifier("do_something"))
    'DoSomething'

    >>> function_name(Identifier("do_something_to_URL"))
    'DoSomethingToUrl'
    """
    return naming.capitalized_camel_case(identifier)


def argument_name(identifier: Identifier) -> Identifier:
    """
    Generate a C++ name for an argument based on its meta-model ``identifier``.

    >>> argument_name(Identifier("something"))
    'something'

    >>> argument_name(Identifier("something_to_URL"))
    'something_to_url'
    """
    return naming.lower_snake_case(identifier)


def variable_name(identifier: Identifier) -> Identifier:
    """
    Generate a C++ name for a variable based on its meta-model ``identifier``.

    >>> variable_name(Identifier("something"))
    'something'

    >>> variable_name(Identifier("something_to_URL"))
    'something_to_url'
    """
    return naming.lower_snake_case(identifier)


def constant_name(identifier: Identifier) -> Identifier:
    """
    Generate a name for a constant based on its meta-model ``identifier``.

    >>> constant_name(Identifier("something"))
    'kSomething'

    >>> constant_name(Identifier("URL_to_something"))
    'kUrlToSomething'
    """
    return Identifier(f"k{naming.capitalized_camel_case(identifier)}")



"""Generate the code to test JSON de/serialization of classes with descendants."""

import io
from typing import List

from icontract import ensure

from aas_core_codegen import intermediate, naming
from aas_core_codegen.common import Stripped, Identifier, indent_but_first_line
from aas_core_codegen.python import common as python_common, naming as python_naming
from aas_core_codegen.python.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
)


def _generate_test_case(symbol_table: intermediate.SymbolTable) -> Stripped:
    test_methods = []  # type: List[Stripped]

    for cls in symbol_table.classes:
        if len(cls.concrete_descendants) == 0:
            continue

        # NOTE (mristin):
        # We can not round-trip instances of classes in JSON which do not store
        # model type.
        if not cls.serialization.with_model_type:
            continue

        test_method_name = python_naming.method_name(Identifier(f"test_{cls.name}"))

        from_jsonable = python_naming.function_name(
            Identifier(f"{cls.name}_from_jsonable")
        )

        descendant_model_types = ",\n".join(
            sorted(
                repr(naming.json_model_type(concrete_cls.name))
                for concrete_cls in cls.concrete_descendants
            )
        )

        test_methods.append(
            Stripped(
                f"""\
def {test_method_name}(self) -> None:
{I}for descendant_model_type in [
{II}{indent_but_first_line(descendant_model_types, II)}
{I}]:
{II}for path in sorted(
{III}(
{IIII}tests.common.TEST_DATA_DIR
{IIII}/ "Json"
{IIII}/ "Expected"
{IIII}/ descendant_model_type
{III}).glob("**/*.json")
{II}):
{III}with path.open("rt") as fid:
{IIII}original_jsonable = json.load(fid)

{III}instance = aas_jsonization.{from_jsonable}(
{IIII}original_jsonable
{III})

{III}another_jsonable = aas_jsonization.to_jsonable(instance)

{III}mismatch = tests.common_jsonization.check_equal(
{IIII}original_jsonable,
{IIII}another_jsonable
{III})

{III}self.assertListEqual([], list(map(str, mismatch)))"""
            ),
        )

    body = (
        "\n\n".join(test_methods)
        if len(test_methods) > 0
        else """\
# There are no classes with concrete descendants.
pass"""
    )

    return Stripped(
        f"""\
class TestRoundTrips(unittest.TestCase):
{I}{indent_but_first_line(body, I)}"""
    )


@ensure(
    lambda result: result.endswith("\n"),
    "Trailing newline mandatory for valid end-of-files",
)
def generate(
    symbol_table: intermediate.SymbolTable,
    qualified_module_name: python_common.QualifiedModuleName,
) -> str:
    """
    Generate the code to test JSON de/serialization of classes with descendants.

    The ``qualified_module_name`` indicates the fully-qualified name of the base module.
    """
    blocks = [
        Stripped(
            '''\
"""Test JSON de/serialization of classes with descendants."""'''
        ),
        python_common.WARNING,
        Stripped(
            """\
# pylint: disable=missing-docstring"""
        ),
        Stripped(
            """\
import json
import unittest"""
        ),
        Stripped(
            f"""\
import {qualified_module_name}.jsonization as aas_jsonization"""
        ),
        Stripped(
            """\
import tests.common_jsonization"""
        ),
        _generate_test_case(symbol_table=symbol_table),
        Stripped(
            f"""\
if __name__ == "__main__":
{I}unittest.main()"""
        ),
        python_common.WARNING,
    ]

    out = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            out.write("\n\n\n")

        out.write(block)

    out.write("\n")

    return out.getvalue()



"""Infer the constraints on a property based on constant sets."""
import collections
from typing import Optional, Tuple, MutableMapping, List, Mapping, Union, Sequence

from icontract import ensure, require

from aas_core_codegen.common import Identifier, Error, assert_never
from aas_core_codegen.infer_for_schema._types import (
    SetOfPrimitivesConstraint,
    SetOfEnumerationLiteralsConstraint,
)
from aas_core_codegen.parse import tree as parse_tree
from aas_core_codegen import intermediate
from aas_core_codegen.infer_for_schema import match as infer_for_schema_match


class _PropNameInNamedContainer:
    """Hold the information for matches of the form ``self.something in X``."""

    def __init__(
        self, prop_name: Identifier, container_name: Identifier, node: parse_tree.IsIn
    ) -> None:
        """Initialize with the given values."""
        self.prop_name = prop_name
        self.container_name = container_name
        self.node = node


def _match_prop_in_named_container(
    node: parse_tree.Node,
) -> Optional[_PropNameInNamedContainer]:
    """Match the expression ``self.something in Something``."""
    if not isinstance(node, parse_tree.IsIn):
        return None

    prop_name = infer_for_schema_match.try_property(node.member)
    if prop_name is None:
        return None

    if not isinstance(node.container, parse_tree.Name):
        return None

    return _PropNameInNamedContainer(
        prop_name=prop_name, container_name=node.container.identifier, node=node
    )


class SetConstraintsByProperty:
    """Group the set constraints by the class properties."""

    def __init__(
        self,
        set_of_primitives_by_property: Mapping[
            intermediate.Property, SetOfPrimitivesConstraint
        ],
        set_of_enumeration_literals_by_property: Mapping[
            intermediate.Property, SetOfEnumerationLiteralsConstraint
        ],
    ) -> None:
        """Initialize with the given values."""
        self.set_of_primitives_by_property = set_of_primitives_by_property
        # fmt: off
        self.set_of_enumeration_literals_by_property = (
                set_of_enumeration_literals_by_property
        )
        # fmt: on


class _IntersectionOfPrimitiveSetLiterals:
    """Track an intersection of primitive set literals."""

    # fmt: off
    @require(
        lambda a_type, literals:
        all(
            literal.a_type is a_type
            for literal in literals
        )
    )
    # fmt: on
    def __init__(
        self,
        a_type: intermediate.PrimitiveType,
        literals: Sequence[intermediate.PrimitiveSetLiteral],
    ) -> None:
        """Initialize with the given set."""
        self.a_type = a_type
        self._literals = literals

        self._count_by_literal_value = {
            literal.value: 0 for literal in literals
        }  # type: MutableMapping[Union[bool, int, float, str, bytearray], int]

        self._counter = 0

    # fmt: off
    @require(
        lambda self, literals:
        all(
            literal.a_type is self.a_type
            for literal in literals
        )
    )
    # fmt: on
    def observe(self, literals: Sequence[intermediate.PrimitiveSetLiteral]) -> None:
        """Limit the intersection by the literals in-place."""
        for literal in literals:
            if literal.value in self._count_by_literal_value:
                self._count_by_literal_value[literal.value] += 1

        self._counter += 1

    def compute_literals(self) -> List[intermediate.PrimitiveSetLiteral]:
        """Compute the literals contained in the intersection."""
        return [
            literal
            for literal in self._literals
            if self._count_by_literal_value[literal.value] == self._counter
        ]


@require(lambda constraints: len(constraints) >= 1)
def intersect_set_of_primitives_constraints(
    constraints: Sequence[SetOfPrimitivesConstraint],
) -> SetOfPrimitivesConstraint:
    """Compute the intersection over all the primitive set literals."""
    intersection = None  # type: Optional[_IntersectionOfPrimitiveSetLiterals]
    for constraint in constraints:
        if intersection is None:
            intersection = _IntersectionOfPrimitiveSetLiterals(
                a_type=constraint.a_type, literals=constraint.literals
            )
        else:
            intersection.observe(literals=constraint.literals)

    assert intersection is not None

    return SetOfPrimitivesConstraint(
        a_type=intersection.a_type, literals=intersection.compute_literals()
    )


class _IntersectionOfEnumerationLiterals:
    """Track an intersection of enumeration literals by their Python object ID."""

    def __init__(self, literals: Sequence[intermediate.EnumerationLiteral]) -> None:
        """Initialize with the given set."""
        self._literals = literals

        self._count_by_literal_id = {
            id(literal): 0 for literal in literals
        }  # type: MutableMapping[int, int]

        self._counter = 0

    def observe(self, literals: Sequence[intermediate.EnumerationLiteral]) -> None:
        """Limit the literals in the intersection."""
        for literal in literals:
            if id(literal) in self._count_by_literal_id:
                self._count_by_literal_id[id(literal)] += 1

        self._counter += 1

    def compute_literals(self) -> List[intermediate.EnumerationLiteral]:
        """Compute the literals contained in the intersection."""
        return [
            literal
            for literal in self._literals
            if self._count_by_literal_id[id(literal)] == self._counter
        ]


# fmt: off
@require(
    lambda constraints:
    all(
        constraint.enumeration is constraints[0].enumeration
        for constraint in constraints
    )
)
@require(
    lambda constraints:
    all(
        all(
            id(literal) in constraint.enumeration.literal_id_set
            for literal in constraint.literals
        )
        for constraint in constraints
    )
)
@require(
    lambda constraints:
    len(constraints) >= 1
)
# fmt: on
def intersect_set_of_enumeration_literals_constraints(
    constraints: Sequence[SetOfEnumerationLiteralsConstraint],
) -> SetOfEnumerationLiteralsConstraint:
    """Compute the intersection over all the set literals."""
    enumeration = None  # type: Optional[intermediate.Enumeration]
    intersection = None  # type: Optional[_IntersectionOfEnumerationLiterals]

    for constraint in constraints:
        if intersection is None:
            enumeration = constraint.enumeration

            intersection = _IntersectionOfEnumerationLiterals(
                literals=constraint.literals
            )
        else:
            intersection.observe(literals=constraint.literals)

    assert enumeration is not None
    assert intersection is not None

    return SetOfEnumerationLiteralsConstraint(
        enumeration=enumeration, literals=intersection.compute_literals()
    )


# fmt: off
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def infer_set_constraints_by_property_from_invariants(
        cls: intermediate.Class,
        symbol_table: intermediate.SymbolTable
) -> Tuple[
    Optional[SetConstraintsByProperty],
    Optional[List[Error]]
]:
    # fmt: on
    """
    Match all the named constant sets that a property needs to belong to.

    Even if a property is optional, the constraint will still be inferred.

    The constraints are not exhaustive. We only infer constraints based on invariants
    which involve constant sets. It might be that the actual invariants are tighter.
    """
    errors = []  # type: List[Error]

    # fmt: off
    sets_of_primitives_by_property: MutableMapping[
        intermediate.Property,
        List[SetOfPrimitivesConstraint]
    ] = collections.defaultdict(lambda: [])
    # fmt: on

    # fmt: off
    sets_of_enumeration_literals_by_property: MutableMapping[
        intermediate.Property, List[SetOfEnumerationLiteralsConstraint]
    ] = collections.defaultdict(lambda: [])
    # fmt: on

    # region Collect

    for invariant in cls.invariants:
        # NOTE (mristin, 2022-07-08):
        # We consider only the genuine invariants of the class, and ignore
        # the invariants of its ancestors.

        if invariant.specified_for is not cls:
            continue

        # Match ``not (self.something is not None) or self.something in X``
        conditional_on_prop = infer_for_schema_match.try_conditional_on_prop(
            invariant.body
        )
        if conditional_on_prop is not None:
            node = conditional_on_prop.consequent
        else:
            # Match ``self.something in X``
            node = invariant.body

        matches = []  # type: List[_PropNameInNamedContainer]

        if isinstance(node, parse_tree.And):
            for value_node in node.values:
                match = _match_prop_in_named_container(value_node)
                if match is not None:
                    matches.append(match)
        else:
            match = _match_prop_in_named_container(node)
            if match is not None:
                matches.append(match)

        for match in matches:
            prop = cls.properties_by_name.get(match.prop_name, None)
            if prop is None:
                errors.append(
                    Error(
                        match.node.member.original_node,
                        f"The property {match.prop_name!r} does not "
                        f"belong to the class {cls.name!r}"
                    )
                )
                continue

            constant = symbol_table.constants_by_name.get(match.container_name, None)
            if constant is None:
                continue

            type_anno = intermediate.beneath_optional(prop.type_annotation)

            if isinstance(constant, intermediate.ConstantPrimitive):
                continue
            elif isinstance(constant, intermediate.ConstantSetOfPrimitives):
                prop_primitive_type = intermediate.try_primitive_type(type_anno)

                if prop_primitive_type is not constant.a_type:
                    errors.append(
                        Error(
                            match.node.container.original_node,
                            f"The container is a constant set "
                            f"of {constant.a_type.value}'s while "
                            f"the property {prop.name!r} in class {cls.name!r} "
                            f"has type {prop.type_annotation}"
                        )
                    )
                    continue

                sets_of_primitives_by_property[prop].append(
                    SetOfPrimitivesConstraint(
                        a_type=constant.a_type,
                        literals=constant.literals
                    )
                )

            elif isinstance(constant, intermediate.ConstantSetOfEnumerationLiterals):
                if not (
                        isinstance(type_anno, intermediate.OurTypeAnnotation)
                        and isinstance(type_anno.our_type, intermediate.Enumeration)
                        and (type_anno.our_type is constant.enumeration)
                ):
                    errors.append(
                        Error(
                            match.node.container.original_node,
                            f"The container is a constant set "
                            f"of enumeration literals of {constant.enumeration.name} "
                            f"while the property {prop.name!r} in class {cls.name!r} "
                            f"has type {prop.type_annotation}"
                        )
                    )
                    continue

                sets_of_enumeration_literals_by_property[prop].append(
                    SetOfEnumerationLiteralsConstraint(
                        enumeration=constant.enumeration,
                        literals=constant.literals
                    )
                )

            else:
                assert_never(constant)

    if len(errors) > 0:
        return None, errors

    # endregion

    # region Reduce

    # fmt: off
    set_of_primitives_by_property: MutableMapping[
        intermediate.Property,
        SetOfPrimitivesConstraint
    ] = dict()
    # fmt: on

    # fmt: off
    set_of_enumeration_literals_by_property: MutableMapping[
        intermediate.Property, SetOfEnumerationLiteralsConstraint
    ] = dict()
    # fmt: on

    for prop, set_of_primitives_constraints in sets_of_primitives_by_property.items():
        set_of_primitives_by_property[prop] = intersect_set_of_primitives_constraints(
            constraints=set_of_primitives_constraints
        )

    for prop, set_of_enumeration_literals_constraints in (
            sets_of_enumeration_literals_by_property.items()
    ):
        set_of_enumeration_literals_by_property[prop] = (
            intersect_set_of_enumeration_literals_constraints(
                constraints=set_of_enumeration_literals_constraints
            )
        )

    # endregion

    return SetConstraintsByProperty(
        set_of_primitives_by_property=set_of_primitives_by_property,
        set_of_enumeration_literals_by_property=set_of_enumeration_literals_by_property
    ), None



"""Provide  structures to model the control flow."""

from typing import Optional, Union, Final, Sequence

from icontract import require

from aas_core_codegen.common import Stripped


class Command:
    """
    Represent a command block to be executed which does not affect control flow.

    We refer to control flow here in terms of yielding. The command block *can* include
    statements changing the *execution* control flow as long as they do not affect
    the co-routines. For example, it can contain an if-statement which does not
    have any yield statements in its body.
    """

    def __init__(self, code: Stripped) -> None:
        """Initialize with the given values."""
        self.code = code


class IfTrue:
    """
    Represent an if-statement in the control flow.

    The condition should *not* contain any code that affects the control flow.
    """

    condition: Final[Stripped]
    body: Final[Sequence["Node"]]
    or_else: Final[Optional[Sequence["Node"]]]

    @require(
        lambda body: len(body) >= 1,
        "If-node should always execute something in the body. If the body is empty, "
        "this node needs to be reformulated",
    )
    def __init__(
        self,
        condition: str,
        body: Sequence["Node"],
        or_else: Optional[Sequence["Node"]] = None,
    ) -> None:
        """Initialize with the given values."""
        self.condition = Stripped(condition.strip())
        self.body = body
        self.or_else = or_else


class IfFalse:
    """
    Represent an if-statement with the negated condition in the control flow.

    The condition should *not* contain any code that affects the control flow.

    We distinguish between if-true and if-false so that we can avoid redundant
    double negations of the condition in the compiled code.
    """

    condition: Final[Stripped]
    body: Final[Sequence["Node"]]
    or_else: Final[Optional[Sequence["Node"]]]

    @require(
        lambda body: len(body) >= 1,
        "If-node should always execute something in the body. If the body is empty, "
        "this node needs to be reformulated",
    )
    def __init__(
        self,
        condition: str,
        body: Sequence["Node"],
        or_else: Optional[Sequence["Node"]] = None,
    ) -> None:
        """Initialize with the given values."""
        self.condition = Stripped(condition.strip())
        self.body = body
        self.or_else = or_else


class For:
    """
    Represent a for-loop in a control flow.

    The ``init``, ``condition`` and ``iteration`` should *not* contain any code that
    affects the control flow.
    """

    init: Final[Optional[Stripped]]
    condition: Final[Stripped]
    iteration: Final[Stripped]
    body: Final[Sequence["Node"]]

    def __init__(
        self,
        condition: str,
        iteration: str,
        body: Sequence["Node"],
        init: Optional[str] = None,
    ) -> None:
        """Initialize with the given values."""
        self.init = Stripped(init) if init is not None else None
        self.condition = Stripped(condition.strip())
        self.iteration = Stripped(iteration.strip())
        self.body = body


class While:
    """Represent a while-loop in a control flow."""

    condition: Final[Stripped]
    body: Final[Sequence["Node"]]

    def __init__(self, condition: str, body: Sequence["Node"]) -> None:
        """Initialize with the given values."""
        self.condition = Stripped(condition.strip())
        self.body = body


class Yield:
    """
    Represent the yield statement.

    Since we want to be as general as possible, we do not include a value to be
    yielded here, as we do not know *how* the compiled code should return the value.
    Therefore, this statement simply indicates that the control is yielded, not
    the value.
    """


Node = Union[Command, IfTrue, IfFalse, For, While, Yield]


def command_from_text(text: str) -> Command:
    """Strip the text and create a command out of the stripped text."""
    return Command(Stripped(text.strip()))



"""Stringify the intermediate representation."""

from typing import Union, Optional

from aas_core_codegen import stringify as stringify_mod
from aas_core_codegen.intermediate import _types, construction
from aas_core_codegen.intermediate._types import (
    AbstractClass,
    Argument,
    Class,
    ConcreteClass,
    ConstrainedPrimitive,
    Constructor,
    Contract,
    Contracts,
    Default,
    DefaultPrimitive,
    DefaultEnumerationLiteral,
    Enumeration,
    EnumerationLiteral,
    ImplementationSpecificMethod,
    ImplementationSpecificVerification,
    Interface,
    Invariant,
    ListTypeAnnotation,
    MetaModel,
    OptionalTypeAnnotation,
    OurTypeAnnotation,
    PatternVerification,
    PrimitiveTypeAnnotation,
    Property,
    Serialization,
    Signature,
    Snapshot,
    OurType,
    SymbolTable,
    UnderstoodMethod,
    DescriptionOfMetaModel,
    DescriptionOfOurType,
    DescriptionOfSignature,
    DescriptionOfProperty,
    DescriptionOfEnumerationLiteral,
    SignatureLike,
    ConstantPrimitive,
    ConstantSetOfPrimitives,
    ConstantSetOfEnumerationLiterals,
    PrimitiveSetLiteral,
    DescriptionOfConstant,
    TranspilableVerification,
)
from aas_core_codegen.parse import tree as parse_tree


def _stringify_primitive_type_annotation(
    that: PrimitiveTypeAnnotation,
) -> stringify_mod.Entity:
    result = stringify_mod.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify_mod.Property("a_type", that.a_type.name),
            stringify_mod.PropertyEllipsis("parsed", that.parsed),
        ],
    )

    return result


def _stringify_our_type_annotation(
    that: OurTypeAnnotation,
) -> stringify_mod.Entity:
    result = stringify_mod.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify_mod.Property(
                "our_type", f"Reference to our type {that.our_type.name}"
            ),
            stringify_mod.PropertyEllipsis("parsed", that.parsed),
        ],
    )

    return result


def _stringify_list_type_annotation(
    that: ListTypeAnnotation,
) -> stringify_mod.Entity:
    result = stringify_mod.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify_mod.Property("items", stringify(that.items)),
            stringify_mod.PropertyEllipsis("parsed", that.parsed),
        ],
    )

    return result


def _stringify_optional_type_annotation(
    that: OptionalTypeAnnotation,
) -> stringify_mod.Entity:
    result = stringify_mod.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify_mod.Property("value", stringify(that.value)),
            stringify_mod.PropertyEllipsis("parsed", that.parsed),
        ],
    )

    return result


def _stringify_description_of_meta_model(
    that: DescriptionOfMetaModel,
) -> stringify_mod.Entity:
    result = stringify_mod.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify_mod.Property("summary", str(that.summary)),
            stringify_mod.Property("remarks", list(map(str, that.remarks))),
            stringify_mod.Property(
                "constraints_by_identifier",
                [
                    [identifier, str(body)]
                    for identifier, body in that.constraints_by_identifier.items()
                ],
            ),
            stringify_mod.PropertyEllipsis("parsed", that.parsed),
        ],
    )

    return result


def _stringify_description_of_our_type(
    that: DescriptionOfOurType,
) -> stringify_mod.Entity:
    result = stringify_mod.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify_mod.Property("summary", str(that.summary)),
            stringify_mod.Property("remarks", list(map(str, that.remarks))),
            stringify_mod.Property(
                "constraints_by_identifier",
                [
                    [identifier, str(body)]
                    for identifier, body in that.constraints_by_identifier.items()
                ],
            ),
            stringify_mod.PropertyEllipsis("parsed", that.parsed),
        ],
    )

    return result


def _stringify_description_of_property(
    that: DescriptionOfProperty,
) -> stringify_mod.Entity:
    result = stringify_mod.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify_mod.Property("summary", str(that.summary)),
            stringify_mod.Property("remarks", list(map(str, that.remarks))),
            stringify_mod.Property(
                "constraints_by_identifier",
                [
                    [identifier, str(body)]
                    for identifier, body in that.constraints_by_identifier.items()
                ],
            ),
            stringify_mod.PropertyEllipsis("parsed", that.parsed),
        ],
    )

    return result


def _stringify_description_of_enumeration_literal(
    that: DescriptionOfEnumerationLiteral,
) -> stringify_mod.Entity:
    result = stringify_mod.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify_mod.Property("summary", str(that.summary)),
            stringify_mod.Property("remarks", list(map(str, that.remarks))),
            stringify_mod.PropertyEllipsis("parsed", that.parsed),
        ],
    )

    return result


def _stringify_description_of_signature(
    that: DescriptionOfSignature,
) -> stringify_mod.Entity:
    result = stringify_mod.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify_mod.Property("summary", str(that.summary)),
            stringify_mod.Property("remarks", list(map(str, that.remarks))),
            stringify_mod.Property(
                "arguments_by_name",
                [[name, str(body)] for name, body in that.arguments_by_name.items()],
            ),
            stringify_mod.Property(
                "returns", None if that.returns is None else str(that.returns)
            ),
            stringify_mod.PropertyEllipsis("parsed", that.parsed),
        ],
    )

    return result


def _stringify_description_of_constant(
    that: DescriptionOfConstant,
) -> stringify_mod.Entity:
    result = stringify_mod.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify_mod.Property("summary", str(that.summary)),
            stringify_mod.Property("remarks", list(map(str, that.remarks))),
            stringify_mod.PropertyEllipsis("parsed", that.parsed),
        ],
    )

    return result


def _stringify_property(
    that: Property,
) -> stringify_mod.Entity:
    result = stringify_mod.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify_mod.Property("name", that.name),
            stringify_mod.Property("type_annotation", stringify(that.type_annotation)),
            stringify_mod.Property("description", stringify(that.description)),
            stringify_mod.Property(
                "specified_for",
                f"Reference to {that.specified_for.__class__.__name__} "
                f"{that.specified_for.name}",
            ),
            stringify_mod.PropertyEllipsis("parsed", that.parsed),
        ],
    )

    return result


def _stringify_default_primitive(
    that: DefaultPrimitive,
) -> stringify_mod.Entity:
    result = stringify_mod.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify_mod.Property("value", that.value),
            stringify_mod.PropertyEllipsis("parsed", that.parsed),
        ],
    )

    return result


def _stringify_default_enumeration_literal(
    that: DefaultEnumerationLiteral,
) -> stringify_mod.Entity:
    result = stringify_mod.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify_mod.Property(
                "enumeration",
                f"Reference to {that.enumeration.__class__.__name__} "
                f"{that.enumeration.name}",
            ),
            stringify_mod.Property(
                "literal",
                f"Reference to {that.literal.__class__.__name__} "
                f"{that.literal.name}",
            ),
        ],
    )

    return result


def _stringify_argument(
    that: Argument,
) -> stringify_mod.Entity:
    result = stringify_mod.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify_mod.Property("name", that.name),
            stringify_mod.Property("type_annotation", stringify(that.type_annotation)),
            stringify_mod.Property("default", stringify(that.default)),
            stringify_mod.PropertyEllipsis("parsed", that.parsed),
        ],
    )

    return result


def _stringify_serialization(
    that: Serialization,
) -> stringify_mod.Entity:
    result = stringify_mod.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify_mod.Property("with_model_type", that.with_model_type),
        ],
    )

    return result


def _stringify_invariant(
    that: Invariant,
) -> stringify_mod.Entity:
    result = stringify_mod.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify_mod.Property("description", that.description),
            stringify_mod.Property("body", parse_tree.dump(that.body)),
            stringify_mod.Property(
                "specified_for",
                f"Reference to {that.specified_for.__class__.__name__} "
                f"{that.specified_for.name}",
            ),
            stringify_mod.PropertyEllipsis("parsed", that.parsed),
        ],
    )

    return result


def _stringify_contract(
    that: Contract,
) -> stringify_mod.Entity:
    result = stringify_mod.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify_mod.Property("args", that.args),
            stringify_mod.Property("description", that.description),
            stringify_mod.Property("body", parse_tree.dump(that.body)),
            stringify_mod.PropertyEllipsis("parsed", that.parsed),
        ],
    )

    return result


def _stringify_snapshot(
    that: Snapshot,
) -> stringify_mod.Entity:
    result = stringify_mod.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify_mod.Property("args", that.args),
            stringify_mod.Property("body", parse_tree.dump(that.body)),
            stringify_mod.Property("name", that.name),
            stringify_mod.PropertyEllipsis("parsed", that.parsed),
        ],
    )

    return result


def _stringify_contracts(that: Contracts) -> stringify_mod.Entity:
    result = stringify_mod.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify_mod.Property(
                "preconditions", list(map(stringify, that.preconditions))
            ),
            stringify_mod.Property("snapshots", list(map(stringify, that.snapshots))),
            stringify_mod.Property(
                "postconditions",
                list(map(stringify, that.postconditions)),
            ),
        ],
    )

    return result


def _stringify_a_signature_like(that: SignatureLike) -> stringify_mod.Entity:
    result = stringify_mod.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify_mod.Property("name", that.name),
            stringify_mod.Property("arguments", list(map(stringify, that.arguments))),
            stringify_mod.Property("returns", stringify(that.returns)),
            stringify_mod.Property("description", stringify(that.description)),
            stringify_mod.Property("contracts", stringify(that.contracts)),
            stringify_mod.PropertyEllipsis("parsed", that.parsed),
            stringify_mod.PropertyEllipsis("arguments_by_name", that.arguments_by_name),
        ],
    )

    return result


def _stringify_implementation_specific_method(
    that: ImplementationSpecificMethod,
) -> stringify_mod.Entity:
    result = stringify_mod.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify_mod.Property("name", that.name),
            stringify_mod.Property("arguments", list(map(stringify, that.arguments))),
            stringify_mod.Property("returns", stringify(that.returns)),
            stringify_mod.Property("description", stringify(that.description)),
            stringify_mod.Property(
                "specified_for",
                f"Reference to {that.specified_for.__class__.__name__} "
                f"{that.specified_for.name}",
            ),
            stringify_mod.Property("contracts", stringify(that.contracts)),
            stringify_mod.Property("non_mutating", that.non_mutating),
            stringify_mod.PropertyEllipsis("parsed", that.parsed),
            stringify_mod.PropertyEllipsis("arguments_by_name", that.arguments_by_name),
        ],
    )

    return result


def _stringify_understood_method(
    that: UnderstoodMethod,
) -> stringify_mod.Entity:
    result = stringify_mod.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify_mod.Property("name", that.name),
            stringify_mod.Property("arguments", list(map(stringify, that.arguments))),
            stringify_mod.Property("returns", stringify(that.returns)),
            stringify_mod.Property("description", stringify(that.description)),
            stringify_mod.Property(
                "specified_for",
                f"Reference to {that.specified_for.__class__.__name__} "
                f"{that.specified_for.name}",
            ),
            stringify_mod.Property("contracts", stringify(that.contracts)),
            stringify_mod.Property("non_mutating", that.non_mutating),
            stringify_mod.PropertyEllipsis("parsed", that.parsed),
            stringify_mod.Property(
                "body", [parse_tree.dump(stmt) for stmt in that.body]
            ),
            stringify_mod.PropertyEllipsis("arguments_by_name", that.arguments_by_name),
        ],
    )

    return result


def _stringify_constructor(
    that: Constructor,
) -> stringify_mod.Entity:
    result = stringify_mod.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify_mod.Property("name", that.name),
            stringify_mod.Property("arguments", list(map(stringify, that.arguments))),
            stringify_mod.Property("returns", stringify(that.returns)),
            stringify_mod.Property("description", stringify(that.description)),
            stringify_mod.Property("contracts", stringify(that.contracts)),
            stringify_mod.PropertyEllipsis("parsed", that.parsed),
            stringify_mod.PropertyEllipsis("arguments_by_name", that.arguments_by_name),
            stringify_mod.Property(
                "is_implementation_specific", that.is_implementation_specific
            ),
            stringify_mod.Property(
                "statements", list(map(construction.dump, that.statements))
            ),
            stringify_mod.Property(
                "inlined_statements",
                list(map(construction.dump, that.inlined_statements)),
            ),
        ],
    )

    return result


def _stringify_enumeration_literal(
    that: EnumerationLiteral,
) -> stringify_mod.Entity:
    result = stringify_mod.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify_mod.Property("name", that.name),
            stringify_mod.Property("value", that.value),
            stringify_mod.Property("description", stringify(that.description)),
            stringify_mod.PropertyEllipsis("parsed", that.parsed),
        ],
    )

    return result


def _stringify_enumeration(
    that: Enumeration,
) -> stringify_mod.Entity:
    result = stringify_mod.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify_mod.Property("name", that.name),
            stringify_mod.Property("literals", list(map(stringify, that.literals))),
            stringify_mod.Property("description", stringify(that.description)),
            stringify_mod.PropertyEllipsis("literals_by_name", that.literals_by_name),
            stringify_mod.PropertyEllipsis("literals_by_value", that.literals_by_value),
            stringify_mod.PropertyEllipsis("literal_id_set", that.literal_id_set),
            stringify_mod.PropertyEllipsis("literal_value_set", that.literal_value_set),
            stringify_mod.PropertyEllipsis("parsed", that.parsed),
        ],
    )

    return result


def _stringify_constrained_primitive(
    that: ConstrainedPrimitive,
) -> stringify_mod.Entity:
    result = stringify_mod.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify_mod.Property("name", that.name),
            stringify_mod.Property(
                "inheritances",
                [
                    f"Reference to {inheritance.__class__.__name__} {inheritance.name}"
                    for inheritance in that.inheritances
                ],
            ),
            stringify_mod.PropertyEllipsis(
                "inheritance_id_set", that.inheritance_id_set
            ),
            stringify_mod.Property(
                "ancestors",
                [
                    f"Reference to {ancestor.__class__.__name__} {ancestor.name}"
                    for ancestor in that.ancestors
                ],
            ),
            stringify_mod.PropertyEllipsis("ancestor_id_set", that.ancestor_id_set),
            stringify_mod.Property(
                "descendants",
                [
                    f"Reference to {descendant.__class__.__name__} {descendant.name}"
                    for descendant in that.descendants
                ],
            ),
            stringify_mod.PropertyEllipsis("descendant_id_set", that.descendant_id_set),
            stringify_mod.Property("constrainee", that.constrainee.name),
            stringify_mod.Property(
                "is_implementation_specific", that.is_implementation_specific
            ),
            stringify_mod.Property("invariants", list(map(stringify, that.invariants))),
            stringify_mod.PropertyEllipsis("invariant_id_set", that.invariant_id_set),
            stringify_mod.Property("description", stringify(that.description)),
            stringify_mod.PropertyEllipsis("parsed", that.parsed),
        ],
    )

    return result


def _stringify_a_class(that: Class) -> stringify_mod.Entity:
    # NOTE (mristin, 2021-12-26):
    # Concrete and abstract class share all the attributes, so we provide a single
    # stringification function.

    result = stringify_mod.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify_mod.Property("name", that.name),
            stringify_mod.Property(
                "inheritances",
                [
                    f"Reference to {inheritance.__class__.__name__} {inheritance.name}"
                    for inheritance in that.inheritances
                ],
            ),
            stringify_mod.PropertyEllipsis(
                "inheritance_id_set", that.inheritance_id_set
            ),
            stringify_mod.Property(
                "ancestors",
                [
                    f"Reference to {ancestor.__class__.__name__} {ancestor.name}"
                    for ancestor in that.ancestors
                ],
            ),
            stringify_mod.PropertyEllipsis("ancestor_id_set", that.ancestor_id_set),
            stringify_mod.Property(
                "is_implementation_specific", that.is_implementation_specific
            ),
            stringify_mod.Property("interface", stringify(that.interface)),
            stringify_mod.PropertyEllipsis("descendant_id_set", that.descendant_id_set),
            stringify_mod.Property(
                "descendants",
                [
                    f"Reference to {descendant.__class__.__name__} {descendant.name}"
                    for descendant in that.descendants
                ],
            ),
            stringify_mod.PropertyEllipsis(
                "concrete_descendant_id_set", that.concrete_descendant_id_set
            ),
            stringify_mod.Property(
                "concrete_descendants",
                [
                    f"Reference to {descendant.__class__.__name__} {descendant.name}"
                    for descendant in that.concrete_descendants
                ],
            ),
            stringify_mod.Property("properties", list(map(stringify, that.properties))),
            stringify_mod.Property("methods", list(map(stringify, that.methods))),
            stringify_mod.Property("constructor", stringify(that.constructor)),
            stringify_mod.Property("invariants", list(map(stringify, that.invariants))),
            stringify_mod.Property("serialization", stringify(that.serialization)),
            stringify_mod.Property("description", stringify(that.description)),
            stringify_mod.PropertyEllipsis("parsed", that.parsed),
            stringify_mod.PropertyEllipsis(
                "properties_by_name", that.properties_by_name
            ),
            stringify_mod.PropertyEllipsis("property_id_set", that.property_id_set),
            stringify_mod.PropertyEllipsis("methods_by_name", that.methods_by_name),
            stringify_mod.PropertyEllipsis("method_id_set", that.method_id_set),
            stringify_mod.PropertyEllipsis("invariant_id_set", that.invariant_id_set),
        ],
    )

    return result


def _stringify_concrete_class(that: ConcreteClass) -> stringify_mod.Entity:
    return _stringify_a_class(that)


def _stringify_abstract_class(that: AbstractClass) -> stringify_mod.Entity:
    return _stringify_a_class(that)


def _stringify_constant_primitive(
    that: ConstantPrimitive,
) -> stringify_mod.Entity:
    result = stringify_mod.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify_mod.Property("name", that.name),
            stringify_mod.Property("value", that.value),
            stringify_mod.Property("a_type", that.a_type.name),
            stringify_mod.Property("description", stringify(that.description)),
            stringify_mod.PropertyEllipsis("parsed", that.parsed),
        ],
    )

    return result


def _stringify_primitive_set_literal(
    that: PrimitiveSetLiteral,
) -> stringify_mod.Entity:
    result = stringify_mod.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify_mod.Property("value", that.value),
            stringify_mod.Property("a_type", that.a_type.name),
            stringify_mod.PropertyEllipsis("parsed", that.parsed),
        ],
    )

    return result


def _stringify_constant_set_of_primitives(
    that: ConstantSetOfPrimitives,
) -> stringify_mod.Entity:
    result = stringify_mod.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify_mod.Property("name", that.name),
            stringify_mod.Property("a_type", that.a_type.name),
            stringify_mod.Property("literals", list(map(stringify, that.literals))),
            stringify_mod.PropertyEllipsis("literal_value_set", that.literal_value_set),
            stringify_mod.Property(
                "subsets",
                [
                    f"Reference to {subset.__class__.__name__} {subset.name}"
                    for subset in that.subsets
                ],
            ),
            stringify_mod.Property("description", stringify(that.description)),
            stringify_mod.PropertyEllipsis("parsed", that.parsed),
        ],
    )

    return result


def _stringify_constant_set_of_enumeration_literals(
    that: ConstantSetOfEnumerationLiterals,
) -> stringify_mod.Entity:
    result = stringify_mod.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify_mod.Property("name", that.name),
            stringify_mod.Property(
                "enumeration",
                f"Reference to {that.enumeration.__class__.__name__} "
                f"{that.enumeration.name}",
            ),
            stringify_mod.Property(
                "literals",
                [
                    f"Reference to {literal.__class__.__name__} {literal.name}"
                    for literal in that.literals
                ],
            ),
            stringify_mod.PropertyEllipsis("literal_id_set", that.literal_id_set),
            stringify_mod.PropertyEllipsis("literal_value_set", that.literal_value_set),
            stringify_mod.Property(
                "subsets",
                [
                    f"Reference to {subset.__class__.__name__} {subset.name}"
                    for subset in that.subsets
                ],
            ),
            stringify_mod.Property("description", stringify(that.description)),
            stringify_mod.PropertyEllipsis("parsed", that.parsed),
        ],
    )

    return result


def _stringify_implementation_specific_verification(
    that: ImplementationSpecificVerification,
) -> stringify_mod.Entity:
    return _stringify_a_signature_like(that)


def _stringify_pattern_verification(
    that: PatternVerification,
) -> stringify_mod.Entity:
    signature_like = _stringify_a_signature_like(that)

    result = stringify_mod.Entity(
        name=that.__class__.__name__,
        properties=list(signature_like.properties)
        + [stringify_mod.Property("pattern", that.pattern)]
        + [stringify_mod.Property("pattern_expr", parse_tree.dump(that.pattern_expr))],
    )

    return result


def _stringify_transpilable_verification(
    that: TranspilableVerification,
) -> stringify_mod.Entity:
    signature_like = _stringify_a_signature_like(that)

    result = stringify_mod.Entity(
        name=that.__class__.__name__, properties=list(signature_like.properties)
    )

    return result


def _stringify_signature(
    that: Signature,
) -> stringify_mod.Entity:
    return _stringify_a_signature_like(that)


def _stringify_interface(
    that: Interface,
) -> stringify_mod.Entity:
    result = stringify_mod.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify_mod.Property(
                "base",
                f"Reference to {that.base.__class__.__name__} {that.base.name}",
            ),
            stringify_mod.Property("name", that.name),
            stringify_mod.Property(
                "inheritances",
                [
                    f"Reference to {inheritance.__class__.__name__} {inheritance.name}"
                    for inheritance in that.inheritances
                ],
            ),
            stringify_mod.Property(
                "implementers",
                [
                    f"Reference to {implementer.__class__.__name__} {implementer.name}"
                    for implementer in that.implementers
                ],
            ),
            stringify_mod.Property("properties", list(map(stringify, that.properties))),
            stringify_mod.Property("signatures", list(map(stringify, that.signatures))),
            stringify_mod.Property("description", stringify(that.description)),
            stringify_mod.PropertyEllipsis("parsed", that.parsed),
            stringify_mod.PropertyEllipsis(
                "properties_by_name", that.properties_by_name
            ),
            stringify_mod.PropertyEllipsis("property_id_set", that.property_id_set),
        ],
    )

    return result


def _stringify_meta_model(
    that: MetaModel,
) -> stringify_mod.Entity:
    result = stringify_mod.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify_mod.Property("description", stringify(that.description)),
            stringify_mod.Property("version", that.version),
            stringify_mod.Property("xml_namespace", that.xml_namespace),
        ],
    )

    return result


def _stringify_symbol_table(
    that: SymbolTable,
) -> stringify_mod.Entity:
    result = stringify_mod.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify_mod.Property("our_types", list(map(stringify, that.our_types))),
            stringify_mod.Property(
                "our_types_topologically_sorted",
                [
                    f"Reference to our type {our_type.name}"
                    for our_type in that.our_types_topologically_sorted
                ],
            ),
            stringify_mod.Property(
                "enumerations",
                [
                    f"Reference to our type {our_type.name}"
                    for our_type in that.enumerations
                ],
            ),
            stringify_mod.Property(
                "constrained_primitives",
                [
                    f"Reference to our type {our_type.name}"
                    for our_type in that.constrained_primitives
                ],
            ),
            stringify_mod.Property(
                "classes",
                [f"Reference to our type {our_type.name}" for our_type in that.classes],
            ),
            stringify_mod.Property(
                "concrete_classes",
                [
                    f"Reference to our type {our_type.name}"
                    for our_type in that.concrete_classes
                ],
            ),
            stringify_mod.Property("constants", list(map(stringify, that.constants))),
            stringify_mod.PropertyEllipsis("constants_by_name", that.constants_by_name),
            stringify_mod.Property(
                "verification_functions",
                list(map(stringify, that.verification_functions)),
            ),
            stringify_mod.PropertyEllipsis(
                "verification_functions_by_name", that.verification_functions_by_name
            ),
            stringify_mod.Property("meta_model", stringify(that.meta_model)),
        ],
    )

    return result


Dumpable = Union[
    AbstractClass,
    Argument,
    ConcreteClass,
    ConstantPrimitive,
    ConstantSetOfEnumerationLiterals,
    ConstantSetOfPrimitives,
    Constructor,
    Contract,
    Contracts,
    Default,
    DescriptionOfMetaModel,
    DescriptionOfOurType,
    DescriptionOfProperty,
    DescriptionOfEnumerationLiteral,
    DescriptionOfSignature,
    DescriptionOfConstant,
    Enumeration,
    EnumerationLiteral,
    ImplementationSpecificMethod,
    ImplementationSpecificVerification,
    Interface,
    Invariant,
    ListTypeAnnotation,
    MetaModel,
    OptionalTypeAnnotation,
    OurTypeAnnotation,
    PatternVerification,
    TranspilableVerification,
    PrimitiveSetLiteral,
    PrimitiveTypeAnnotation,
    Property,
    Serialization,
    Signature,
    Snapshot,
    OurType,
    SymbolTable,
    UnderstoodMethod,
]

stringify_mod.assert_all_public_types_listed_as_dumpables(
    dumpable=Dumpable, types_module=_types
)

_DISPATCH = {
    AbstractClass: _stringify_abstract_class,
    Argument: _stringify_argument,
    ConcreteClass: _stringify_concrete_class,
    ConstantPrimitive: _stringify_constant_primitive,
    ConstantSetOfEnumerationLiterals: _stringify_constant_set_of_enumeration_literals,
    ConstantSetOfPrimitives: _stringify_constant_set_of_primitives,
    ConstrainedPrimitive: _stringify_constrained_primitive,
    Constructor: _stringify_constructor,
    Contract: _stringify_contract,
    Contracts: _stringify_contracts,
    DefaultPrimitive: _stringify_default_primitive,
    DefaultEnumerationLiteral: _stringify_default_enumeration_literal,
    DescriptionOfConstant: _stringify_description_of_constant,
    DescriptionOfMetaModel: _stringify_description_of_meta_model,
    DescriptionOfOurType: _stringify_description_of_our_type,
    DescriptionOfProperty: _stringify_description_of_property,
    DescriptionOfEnumerationLiteral: _stringify_description_of_enumeration_literal,
    DescriptionOfSignature: _stringify_description_of_signature,
    Enumeration: _stringify_enumeration,
    EnumerationLiteral: _stringify_enumeration_literal,
    ImplementationSpecificMethod: _stringify_implementation_specific_method,
    ImplementationSpecificVerification: _stringify_implementation_specific_verification,
    Interface: _stringify_interface,
    Invariant: _stringify_invariant,
    ListTypeAnnotation: _stringify_list_type_annotation,
    MetaModel: _stringify_meta_model,
    OptionalTypeAnnotation: _stringify_optional_type_annotation,
    OurTypeAnnotation: _stringify_our_type_annotation,
    PatternVerification: _stringify_pattern_verification,
    PrimitiveSetLiteral: _stringify_primitive_set_literal,
    PrimitiveTypeAnnotation: _stringify_primitive_type_annotation,
    Property: _stringify_property,
    Serialization: _stringify_serialization,
    Signature: _stringify_signature,
    Snapshot: _stringify_snapshot,
    SymbolTable: _stringify_symbol_table,
    TranspilableVerification: _stringify_transpilable_verification,
    UnderstoodMethod: _stringify_understood_method,
}

stringify_mod.assert_dispatch_exhaustive(dispatch=_DISPATCH, dumpable=Dumpable)


def stringify(that: Optional[Dumpable]) -> Optional[stringify_mod.Entity]:
    """Dispatch to the correct ``_stringify_*`` method."""
    if that is None:
        return None

    stringify_func = _DISPATCH.get(that.__class__, None)
    if stringify_func is None:
        raise AssertionError(
            f"No stringify function could be found for the class {that.__class__}"
        )

    stringified = stringify_func(that)  # type: ignore
    assert isinstance(stringified, stringify_mod.Entity)
    stringify_mod.assert_compares_against_dict(stringified, that)

    return stringified


def dump(that: Optional[Dumpable]) -> str:
    """Produce a string representation of the ``dumpable`` for testing or debugging."""
    if that is None:
        return repr(None)

    stringified = stringify(that)
    return stringify_mod.dump(stringified)



"""Generate the test code for the xmlization errors."""

from typing import List

from icontract import ensure

from aas_core_codegen import intermediate, naming
from aas_core_codegen.common import Stripped, indent_but_first_line, Identifier
from aas_core_codegen.csharp import common as csharp_common, naming as csharp_naming
from aas_core_codegen.csharp.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
)


# fmt: off
@ensure(
    lambda result: result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(
    namespace: csharp_common.NamespaceIdentifier,
    symbol_table: intermediate.SymbolTable,
) -> str:
    """
    Generate the test code for the xmlization errors.

    The ``namespace`` indicates the fully-qualified name of the base project.
    """
    if len(symbol_table.concrete_classes) == 0:
        raise ValueError(
            "Expected at least one concrete class in the symbol table, but got none."
        )

    concrete_cls = symbol_table.concrete_classes[0]

    cls_name_xml = naming.xml_class_name(concrete_cls.name)

    from_name = csharp_naming.method_name(Identifier(f"{concrete_cls.name}_from"))

    cls_name_csharp = csharp_naming.class_name(concrete_cls.name)

    blocks = [
        Stripped(
            f"""\
[Test]
public void Test_error_on_unexpected_declaration()
{{
{I}string path = Path.Combine(
{II}Aas.Tests.Common.TestDataDir,
{II}"Xml",
{II}"Expected",
{II}{csharp_common.string_literal(cls_name_xml)},
{II}"minimal.xml");

{I}var text = System.IO.File.ReadAllText(path, System.Text.Encoding.UTF8);

{I}// We assume no XML declarations in our test data.
{I}if (!text.StartsWith("<{cls_name_xml} "))
{I}{{
{II}throw new System.InvalidOperationException(
{III}"We expect our test example in XML to start with '<{cls_name_xml} ', "
{IIII}+ $"but it does not: {{path}}"
{II});
{I}}}

{I}text = (
{II}"<?xml version=\\"1.0\\" encoding=\\"utf-8\\"?>"
{III}+ text
{I});

{I}using var stringReader = new System.IO.StringReader(
{II}text);

{I}using var xmlReader = System.Xml.XmlReader.Create(
{II}stringReader);

{I}// We intentionally do not call `MoveToContent` to test the error message.
{I}// This is a very common situation, see:
{I}// https://github.com/aas-core-works/aas-core3.0-csharp/issues/24

{I}string? message = null;

{I}try
{I}{{
{II}Aas.Xmlization.Deserialize.{from_name}(
{III}xmlReader);
{I}}}
{I}catch (Aas.Xmlization.Exception exception)
{I}{{
{II}message = exception.Message;
{I}}}

{I}if (message == null)
{I}{{
{II}throw new AssertionException("Unexpected no exception");
{I}}}
{I}Assert.AreEqual(
{II}"Unexpected XML declaration when reading an instance of " +
{II}"class {cls_name_csharp}, as we expect the reader to be set at content " +
{II}"with MoveToContent at: the beginning",
{II}message);
}}"""
        ),
    ]  # type: List[Stripped]

    blocks_joined = "\n\n".join(blocks)

    return f"""\
{csharp_common.WARNING}

using Aas = {namespace};  // renamed

using Path = System.IO.Path;

using NUnit.Framework; // can't alias

namespace {namespace}.Tests
{{
{I}public class TestXmlizationErrors
{I}{{
{II}{indent_but_first_line(blocks_joined, II)}
{I}}}  // class TestXmlizationErrors
}}  // namespace {namespace}.Tests

{csharp_common.WARNING}
"""


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())
